import sys
import os

# Should all edit descriptor values be returned even if they were not
# written to?
RET_WRITTEN_VARS_ONLY = False
# Should 'None' values be returned when no record is available to read
# from or the FORTRAN 'default'?
RET_UNWRITTEN_VARS_NONE = True
# The order in which edit desciptors are tried by default when G edit
# descriptor encountered on input
G_INPUT_TRIAL_EDS = ['F', 'L', 'A']
# Contrary to specification, many compilers allow zero width edit
# descriptors
ALLOW_ZERO_WIDTH_EDS = True
# Set the characters that separate the records
RECORD_SEPARATOR = '\n'

# The maximum size for an integer
if sys.version_info[0] >= 3:
    PROC_MAXINT = sys.maxsize
else:
    PROC_MAXINT = sys.maxint
# Processor dependant default for including leading plus or not
PROC_INCL_PLUS = False 
# Option to allow signed binary, octal and hex on input (not a FORTRAN feature)
PROC_ALLOW_NEG_BOZ = False
# Prcessor dependant padding character
PROC_PAD_CHAR = ' '
# Interpret blanks or jsut a negative as a zero, as in ifort behaviour
PROC_NEG_AS_ZERO = True
# Show a sign for zero?
PROC_SIGN_ZERO = False
PROC_MIN_FIELD_WIDTH = 46
PROC_DECIMAL_CHAR = '.'
G0_NO_BLANKS = False
PROC_NO_LEADING_BLANK = False
# The default value if BN, BZ edit descriptors are not specified
PROC_BLANKS_AS_ZEROS = False

def reset():
    global RET_WRITTEN_VARS_ONLY, RET_UNWRITTEN_VARS_NONE, PROC_INCL_PLUS, \
        PROC_ALLOW_NEG_BOZ, PROC_PAD_CHAR, PROC_NEG_AS_ZERO, PROC_SIGN_ZERO, \
        PROC_MIN_FIELD_WIDTH, PROC_DECIMAL_CHAR, G0_NO_BLANKS, \
        PROC_NO_LEADING_BLANK, PROC_BLANKS_AS_ZEROS, PROC_MAXINT, G_INPUT_TRIAL_EDS, \
        ALLOW_ZERO_WIDTH_EDS
    G_INPUT_TRIAL_EDS = ['F', 'L', 'A']
    if sys.version_info[0] >= 3:
        PROC_MAXINT = sys.maxsize
    else:
        PROC_MAXINT = sys.maxint
    RET_WRITTEN_VARS_ONLY = False
    RET_UNWRITTEN_VARS_NONE = True
    PROC_INCL_PLUS = False
    PROC_ALLOW_NEG_BOZ = False
    PROC_PAD_CHAR = ' '
    PROC_NEG_AS_ZERO = True
    PROC_SIGN_ZERO = False
    PROC_MIN_FIELD_WIDTH = 46
    PROC_DECIMAL_CHAR = '.'
    G0_NO_BLANKS = False
    PROC_NO_LEADING_BLANK = False
    PROC_BLANKS_AS_ZEROS = False
    ALLOW_ZERO_WIDTH_EDS = True



""" Package that provides an API to Amber-specific files """

__author__ = "Jason Swails <jason.swails@gmail.com>"

__all__ = ['AmberFormat', 'FortranFormat', 'AmberAsciiRestart', 'AmberMdcrd',
           'AmberMask', 'NetCDFTraj', 'NetCDFRestart', 'AmberOFFLibrary',
           'AmberParameterSet', 'AmberParm', 'ChamberParm', 'AmoebaParm',
           'Rst7', 'BeemanRestart', 'ConvertFromPSF', 'LoadParm', 'AMBERHOME',
           'titratable_residues']

from parmed.amber.amberformat import AmberFormat, FortranFormat
from parmed.amber.asciicrd import AmberAsciiRestart, AmberMdcrd
from parmed.amber.mask import AmberMask
from parmed.amber.netcdffiles import NetCDFTraj, NetCDFRestart
from parmed.amber.offlib import AmberOFFLibrary
from parmed.amber.parameters import AmberParameterSet
from parmed.amber.readparm import (AmberParm, ChamberParm, AmoebaParm,
                Rst7, BeemanRestart, ConvertFromPSF, LoadParm)
from parmed.amber import titratable_residues

# See if there is an AMBERHOME defined, which we will use by default. Otherwise,
# set it to the empty string
import os as _os
AMBERHOME = _os.getenv('AMBERHOME') or ''
del _os



"""
This module contains all of the methods to invoke specific ParmEd actions listed
in that module. Each method will establish the necessary variables (StringVar),
create whatever window it needs to get information from the user (or give
information to the user), wait for that window to close, then dispatch that
Action to the class in ParmEd actions.

Follow the general trend if you wish to add your method to the GUI. Note, any
method that you want accessible through the GUI must have an action method put
here with the same name as the class found in ParmEd actions.
"""

from parmed.utils.six.moves import range
from parmed.utils.six.moves.tkinter import *
from parmed.utils.six.moves.tkinter_messagebox import (
        askyesno, showinfo, showerror)
from parmed.tools import actions
from parmed.tools.argumentlist import ArgumentList
from parmed.tools.gui.guifiletools import file_chooser, save_file_chooser
from parmed.tools.gui import _guiwidgets

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def writefrcmod(root, amber_prmtop, messages):
    """ Dumps an frcmod file to a given filename """
    fname = save_file_chooser('frcmod', '.frcmod')
    if fname: 
        action = actions.writeFrcmod(amber_prmtop, fname)
        action.execute()
        messages.write('%s\n' % action)

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def loadrestrt(root, amber_prmtop, messages):
    """ Finds a file to load as a restart """
    type_list = [('Inpcrd', '*.inpcrd'), ('Inpcrd', '*.crd'),
                 ('Restart', '*.restrt'), ('Restart', '*.rst7'),
                 ('All Files', '*')]
    fname = file_chooser('Amber Coordinate File', type_list)
    if fname: 
        action = actions.loadRestrt(amber_prmtop, ArgumentList(fname))
        messages.write('%s\n' % action)
        action.execute()

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def writeoff(root, amber_prmtop, messages):
    """ Dumps an OFF library to a given filename """
    fname = save_file_chooser('OFF library', '.lib')
    if fname: 
        action = actions.writeOFF(amber_prmtop, ArgumentList(fname))
        messages.write('%s\n' % action)
        action.execute()

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def changeradii(root, amber_prmtop, messages):
    """ Allows users to change the GB Radius set """
    title = 'Choose a Radius Set'
    desc = ('Select a radius set for implicit solvent\n' +
            'calculations. This has the same effect as\n' +
            '"set default PBRadii <value>" in tleap')
    radius_selection = StringVar()
    namelist = ['bondi', 'mbondi', 'mbondi2', 'mbondi3', 'amber6']
    cmd_window = _guiwidgets.RadioButtonWindow(
                    amber_prmtop, title, desc, radius_selection, namelist)
    # Wait until the window is destroyed, then get the variable and pass it
    # over to the class
    cmd_window.wait_window()
    sel = str(radius_selection.get())
    if sel:
        action = actions.changeRadii(amber_prmtop, ArgumentList(sel))
        messages.write('%s\n' % action)
        action.execute()

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def changeljpair(root, amber_prmtop, messages):
    """ Changes a pair-wise LJ interaction for pre-combined epsilon/Rmin """
    # The variables we need for changeljpair
    widget_list = [('MaskEntry', 'Atom(s) Type 1 Mask'),
                   ('MaskEntry', 'Atom(s) Type 2 Mask'),
                   ('Entry', 'Combined Radius'),
                   ('Entry', 'Combined Well Depth')]
    # Variable list -- we need 2 masks and 2 floats
    var_list = [StringVar(), StringVar(), StringVar(), StringVar()]
    # description
    description = ' '.join(actions.changeLJPair.__doc__.split())
    cmd_window = _guiwidgets.ActionWindow('changeLJPair', amber_prmtop,
                                widget_list, var_list, description)
    cmd_window.wait_window()
    # Make sure we didn't cancel (or just press OK with no input), or just leave
    vars_exist = True in [bool(v.get()) for v in var_list]
    if not vars_exist: return
    # Now that we did something, do it
    var_list = [v.get() for v in var_list]
    try:
        action = actions.changeLJPair(amber_prmtop,ArgumentList(var_list))
        messages.write('%s\n' % action)
        action.execute()
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def outparm(root, amber_prmtop, messages):
    """ Output a final topology file """
    fname = [save_file_chooser('prmtop', '.prmtop')]
    if amber_prmtop.parm.coords is not None and fname[0]:
        fname.append(save_file_chooser('inpcrd', '.inpcrd'))
    if fname[0]:
        action = actions.outparm(amber_prmtop, ArgumentList(fname))
        messages.write('%s\n' % action)
        action.execute()

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def printflags(root, amber_prmtop, messages):
    """ Prints all of the flags in the topology file """
    window = Toplevel(root)
    window.resizable(True, True)
    window.title('%%FLAG list in %s' % amber_prmtop.parm)
    text = _guiwidgets.ExitingScrollText(window, None, spacing3=5, padx=5,
                                         pady=5, width=80, height=20)
    text.pack(fill=BOTH, expand=1)
    action = actions.printFlags(amber_prmtop, ArgumentList(''))
    text.write(action)
    window.wait_window()

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def printpointers(root, amber_prmtop, messages):
    """ Prints all of the flags in the topology file """
    window = Toplevel(root)
    window.resizable(True, True)
    window.title('POINTER list in %s' % amber_prmtop.parm)
    text = _guiwidgets.ExitingScrollText(window, None, spacing3=5, padx=5,
                                         pady=5, width=80, height=20)
    text.pack(fill=BOTH, expand=1)
    action = actions.printPointers(amber_prmtop, ArgumentList(''))
    text.write(action)
    window.wait_window()

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def changelj14pair(root, amber_prmtop, messages):
    """ Changes specific 1-4 Lennard Jones pairs """
    # Only good for chamber topologies
    if not amber_prmtop.parm.chamber:
        showerror('Incompatible',
                  'changeLJ14Pair is only valid for chamber topologies!')
        return
    # variables we need for changelj14pair
    widget_list = [('MaskEntry', 'Atom(s) Type 1 Mask'),
                   ('MaskEntry', 'Atom(s) Type 2 Mask'),
                   ('Entry', 'Combined Radius'),
                   ('Entry', 'Combined Well Depth')]
    # Variable list -- we need 2 masks and 2 floats
    var_list = [StringVar(), StringVar(), StringVar(), StringVar()]
    # description
    description = ' '.join(actions.changeLJ14Pair.__doc__.split())
    cmd_window = _guiwidgets.ActionWindow('changeLJ14Pair', amber_prmtop,
                                widget_list, var_list, description)
    cmd_window.wait_window()
    # Make sure we didn't cancel (or just press OK with no input), or just leave
    vars_exist = True in [bool(v.get()) for v in var_list]
    if not vars_exist: return
    # Now that we did something, do it
    var_list = [v.get() for v in var_list]
    try:
        action=actions.changeLJ14Pair(amber_prmtop,ArgumentList(var_list))
        messages.write('%s\n' % action)
        action.execute()
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def checkvalidity(root, amber_prmtop, messages):
    """ Basic validity checks """
    # Create our Info window
    window = Toplevel(root)
    window.resizable(True, True)
    window.title('Problems with %s' % amber_prmtop.parm)
    text = _guiwidgets.ExitingScrollText(window, None, spacing3=2, padx=5,
                                         pady=5, width=81, height=30)
    text.pack(fill=BOTH, expand=1)
    # Set this text to catch the output of our action
    actions.checkValidity.output = text
    # Initialize our action
    action = actions.checkValidity(amber_prmtop, ArgumentList(''))
    messages.write('%s\n' % action)
    action.execute()
    text.write(action)

    window.wait_window()

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def change(root, amber_prmtop, messages):
    """ Allows us to change a specific atomic property """
    # The spinbox is sent with the Spinbox, label, and then a list of all of the
    # values to give to it
    widget_list = [('Spinbox', 'Property to change', 'CHARGE', 'MASS',
                    'RADII', 'SCREEN', 'ATOM_NAME', 'AMBER_ATOM_TYPE',
                    'ATOM_TYPE_INDEX', 'ATOMIC_NUMBER'),
                    ('MaskEntry', 'Atoms to change'),
                    ('Entry', 'New Value for Property')]
    # We need 3 string variables, then get the description
    var_list = [StringVar(), StringVar(), StringVar()]
    description = 'Changes the property of given atoms to a new value'
    # Create the window, open it, then wait for it to close
    cmd_window = _guiwidgets.ActionWindow('change', amber_prmtop,
                        widget_list, var_list, description)
    cmd_window.wait_window()
    # See if we got any variables back
    vars_found = True in [bool(v.get()) for v in var_list]
    if not vars_found: return
    # If we did, store them and pass it to the class
    var_list = [v.get() for v in var_list]
    try:
        action = actions.change(amber_prmtop, ArgumentList(var_list))
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))
        return
    action.execute()
    messages.write('%s\n' % action)

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def printinfo(root, amber_prmtop, messages):
    """ Prints all of the info in a given FLAG """
    # Set up the window
    # variables we need for printInfo
    widget_list = [('Entry', '%FLAG you want info from')]
    # Variable list -- we need a single string
    var_list = [StringVar()]
    # description
    description = ' '.join(actions.printInfo.__doc__.split())
    cmd_window = _guiwidgets.ActionWindow('printInfo', amber_prmtop,
                                widget_list, var_list, description)
    cmd_window.wait_window()
    # Make sure we didn't cancel (or just press OK with no input), or just leave
    var = var_list[0].get()
    if not var: return
    # Now that we did something, do it
    action = actions.printInfo(amber_prmtop, ArgumentList(var))
    if not action.found:
        showerror('Not Found!', '%%FLAG %s not found!' % var.upper())
        return

    window = Toplevel(root)
    window.resizable(True, True)
    window.title('%%FLAG %s Info in %s' % (var.upper(), amber_prmtop))
    text = _guiwidgets.ExitingScrollText(window, None, spacing3=5, padx=5,
                                            pady=5, width=82, height=20)
    text.write(action)
    text.pack(fill=BOTH, expand=1)
    window.wait_window()

    messages.write('Wrote info for flag %s\n' % var.upper())

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def addljtype(root, amber_prmtop, messages):
    """ Turns given mask into a new LJ atom type """
    # We need a mask, new radius, new epsilon, and for chamber topologies,
    # a new radius-1-4 and epsilon-1-4. Don't add the latter ones until we
    # know if we have a chamber prmtop or not.
    widget_list = [('MaskEntry', 'Atoms to make new LJ Type'),
                   ('Entry', 'New Radius (default old radius)'),
                   ('Entry', 'New Depth (default old depth)')]
    # We need 5 string variables, then get the description. 
    var_list = [StringVar(), StringVar(), StringVar(), StringVar(), StringVar()]
    description=('Turns given mask into a new LJ atom type. Uses the radius\n'
                 'and well depth from the first atom type in <mask> if none\n'
                 'are provided.')
    if amber_prmtop.parm.chamber:
        widget_list += [('Entry', 'New Radius for 1-4 Terms'),
                        ('Entry', 'New Depth for 1-4 Terms')]
    # Create the window, open it, then wait for it to close
    cmd_window = _guiwidgets.ActionWindow('addLJType', amber_prmtop,
                        widget_list, var_list, description)
    cmd_window.wait_window()
    # See if we got any variables back
    vars_found = True in [bool(v.get()) for v in var_list]
    if not vars_found: return
    # addljtype expects any _non_specified variables to be None
    var_list = [v.get().strip() for v in var_list]
    kw_var_list = [var_list[0]]
    if var_list[1]: kw_var_list.extend(['radius', var_list[1]])
    if var_list[2]: kw_var_list.extend(['epsilon', var_list[2]])
    if amber_prmtop.parm.chamber and var_list[3]:
        kw_var_list.extend(['radius_14', var_list[3]])
    if amber_prmtop.parm.chamber and var_list[4]:
        kw_var_list.extend(['epsilon_14', var_list[4]])
    try:
        action = actions.addLJType(amber_prmtop,ArgumentList(kw_var_list))
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))
        return
    action.execute()
    messages.write('%s\n' % action)

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def setmolecules(root, amber_prmtop, messages):
    """ Sets the molecules. Asks users if they want ions to be solute or not """
    title = 'Set Molecularity'
    desc = ('This will determine the molecular topology (ATOMS_PER_MOLECULE\n'
            'and SOLVENT_POINTERS). Do you want the ions to be considered\n'
            'part of the solute?')
    solute_ions = StringVar()
    namelist = ['Yes', 'No']
    cmd_window = _guiwidgets.RadioButtonWindow(
                            amber_prmtop, title, desc, solute_ions, namelist)
    # Wait until the window is destroyed, then get the variable and pass it
    # over to the class
    cmd_window.wait_window()
    sel = str(solute_ions.get())
    if not sel: return

    if sel == 'Yes':
        sel = 'True'
    else:
        sel = 'False'
    action = actions.setMolecules(amber_prmtop, 
                                        ArgumentList('solute_ions '+sel))
    messages.write('%s\n' % action)
    action.execute()

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def printdetails(root, amber_prmtop, messages):
    """ Prints details about a given Amber mask """
    title = 'Print Details'
    mask = StringVar()
    cmd_window = Toplevel(root)
    cmd_window.title(title)
    mask_entry = _guiwidgets.MaskEntry(cmd_window, amber_prmtop,
                                'Input an Amber Mask', mask, cmd_window)
    mask_entry.config(pady=10)
    mask_entry.grid(row=0, column=0, sticky=N+E+S+W)
    button = Button(cmd_window, text='OK / Quit', command=cmd_window.destroy)
    button.grid(row=1, column=0, sticky=N+E+S+W)
    cmd_window.wait_window()
    if not mask.get(): return
    # Print our mask
    window = Toplevel(root)
    window.resizable(True, True)
    window.title('Atom information for mask %s' % mask.get())
    text = _guiwidgets.ExitingScrollText(window, None, spacing3=5, padx=5,
                                         pady=5, width=100, height=20)
    text.pack(fill=BOTH, expand=1)
    action = actions.printDetails(amber_prmtop, mask.get())
    text.write(action)
    messages.write('Printed Amber Mask details on [%s]\n' % mask.get())
    window.wait_window()

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def netcharge(root, amber_prmtop, messages):
    """ Calculates the net charge and shows its value """
    mask = StringVar()
    cmd_window = Toplevel(root)
    mask_entry = _guiwidgets.MaskEntry(cmd_window, amber_prmtop,
                        'Mask from which to calculate charge', mask, cmd_window)
    mask_entry.config(pady=10)
    mask_entry.grid(row=0, column=0, sticky=N+E+S+W)
    button = Button(cmd_window, text='OK / Quit', command=cmd_window.destroy)
    button.grid(row=1, column=0, sticky=N+E+S+W)
    cmd_window.wait_window()
    if not mask.get(): return
    action = actions.netCharge(amber_prmtop, ArgumentList(mask.get()))
    chg = action.execute()
    showinfo('Net Charge', 'The net charge of [%s] is %.4f' % (mask.get(), chg))

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def strip(root, amber_prmtop, messages):
    """ Strips a mask from the topology file """
    # We need a mask, new radius, new epsilon, and for chamber topologies,
    # a new radius-1-4 and epsilon-1-4. Don't add the latter ones until we
    # know if we have a chamber prmtop or not.
    widget_list = [('MaskEntry', 'Atoms to strip from topology')]
    # We need 5 string variables, then get the description. 
    var_list = [StringVar()]
    description=('Strips the selected atoms from the topology file. All\n'
                 'remaining atoms and parameters remain unchanged. Any\n'
                 'parameters associated with stripped atoms are removed.')
    # Create the window, open it, then wait for it to close
    cmd_window = _guiwidgets.ActionWindow('strip', amber_prmtop,
                                        widget_list, var_list, description)
    cmd_window.wait_window()
    # See if we got any variables back
    var = var_list[0]
    if not var.get(): return
    try:
        action = actions.strip(amber_prmtop, ArgumentList(var.get()))
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))
        return
    messages.write('%s\n' % action)
    action.execute()

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def addexclusions(root, amber_prmtop, messages):
    """ Adds atoms to other atoms' exclusion list """
    # We need 2 masks
    widget_list = [('MaskEntry', 'Atoms to add excluded atoms to'),
                   ('MaskEntry', 'Atoms to exclude from other mask')]
    # We have 2 mask variables
    var_list = [StringVar(), StringVar()]
    # Description
    description = ('Allows you to add arbitrary excluded atoms to exclusion\n'
                   'lists. This omits all non-bonded interactions in the '
                   'direct-space\ncalculation, but does not omit interactions '
                   'from adjacent cells in\nperiodic simulations')
    cmd_window = _guiwidgets.ActionWindow('addExclusions', amber_prmtop,
                                          widget_list, var_list, description)
    cmd_window.wait_window()
   
    # Bail out if we didn't get any variables
    if not True in [bool(v.get()) for v in var_list]: return
   
    var_list = [v.get() for v in var_list]
    try:
        act = actions.addExclusions(amber_prmtop, ArgumentList(var_list))
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))
        return
    act.execute()
    messages.write('%s\n' % act)

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def changeprotstate(root, amber_prmtop, messages):
    # We need a mask and a state
    widget_list = [('MaskEntry', 'Residue to change protonation state'),
                   ('Entry', 'Protonation state to change to')]
    var_list = [StringVar(), StringVar()]
    description=('Changes the protonation state of a pH-active titratable residue that\n'
                 'can be treated with constant pH MD in Amber.')
    cmd_window = _guiwidgets.ActionWindow('changeProtState', amber_prmtop,
                        widget_list, var_list, description)
    cmd_window.wait_window()
   
    # Bail out if we didn't get any variables
    if not True in [bool(v.get()) for v in var_list]: return
   
    var_list = [v.get() for v in var_list]
    try:
        action = actions.changeProtState(amber_prmtop, 
                                               ArgumentList(var_list))
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))
        return
    action.execute()
    messages.write('%s\n' % action)

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def changeredustate(root, amber_prmtop, messages):
    # We need a mask and a state
    widget_list = [('MaskEntry', 'Residue to change reduction state'),
                   ('Entry', 'Reduction state to change to')]
    var_list = [StringVar(), StringVar()]
    description=('Changes the reduction state of a redox-active titratable residue that\n'
                 'can be treated with constant redox potential MD in Amber.')
    cmd_window = _guiwidgets.ActionWindow('changeRedoxState', amber_prmtop,
                        widget_list, var_list, description)
    cmd_window.wait_window()
   
    # Bail out if we didn't get any variables
    if not True in [bool(v.get()) for v in var_list]: return
   
    var_list = [v.get() for v in var_list]
    try:
        action = actions.changeRedoxState(amber_prmtop, 
                                               ArgumentList(var_list))
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))
        return
    action.execute()
    messages.write('%s\n' % action)

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def scee(root, amber_prmtop, messages):
    # We need a value
    widget_list = [('Entry', '1-4 Electrostatic Scaling Factor')]
    var_list = [StringVar()]
    description = 'Adjust the scaling factor for 1-4 electrostatic interactions'

    cmd_window = _guiwidgets.ActionWindow('scee', amber_prmtop,
                                          widget_list, var_list, description)
    cmd_window.wait_window()
   
    var = var_list[0].get()

    # Bail out if we didn't get any variables
    if not var: return

    try:
        action = actions.scee(amber_prmtop, ArgumentList(var))
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))
        return
    messages.write('%s\n' % action)
    action.execute()

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def scnb(root, amber_prmtop, messages):
    # We need a value
    widget_list = [('Entry', '1-4 van der Waals Scaling Factor')]
    var_list = [StringVar()]
    description = 'Adjust the scaling factor for 1-4 van der Waals interactions'

    cmd_window = _guiwidgets.ActionWindow('scee', amber_prmtop,
                                          widget_list, var_list, description)
    cmd_window.wait_window()
   
    var = var_list[0].get()

    # Bail out if we didn't get any variables
    if not var: return

    try:
        action = actions.scnb(amber_prmtop, ArgumentList(var))
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))
        return
    messages.write('%s\n' % action)
    action.execute()

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def printljtypes(root, amber_prmtop, messages):
    """
    Prints all of the atoms that have the same LJ type as atoms in a given mask
    """
    # We need a mask
    widget_list = [('MaskEntry', 'Atom Mask or Atom Type Name')]
    var_list = [StringVar()]
    description = ('Get a list of all atoms that share a common LJ type')
    cmd_window = _guiwidgets.ActionWindow('printLJTypes', amber_prmtop,
                                          widget_list, var_list, description)
    cmd_window.wait_window()
   
    var = var_list[0].get()

    # Bail out if we didn't get any variables
    if not var: return
   
    # Instantiate our action
    try:
        action = actions.printLJTypes(amber_prmtop, ArgumentList(var))
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))
        return

    # Create our Info window
    window = Toplevel(root)
    window.resizable(True, True)
    window.title('LJ Type List')
    text = _guiwidgets.ExitingScrollText(window, None, spacing3=2, padx=5,
                                         pady=5, width=45, height=30)
    text.pack(fill=BOTH, expand=1)
    text.write(action)
    messages.write('Printed LJ types for [%s]\n' % var)
    window.wait_window()

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def changeljsingletype(root, amber_prmtop, messages):
    """ Changes radius/well depth of a single LJ type given by the mask """
    # We need a mask, radius, and well depth
    widget_list = [('MaskEntry', 'Mask to change LJ Type'),
                   ('Entry', 'New LJ Radius'), ('Entry', 'New LJ Depth')]
    var_list = [StringVar(), StringVar(), StringVar()]
    description = "Change a given atom type's LJ radius and well depth"
    # Create the window, open it, then wait for it to close
    cmd_window = _guiwidgets.ActionWindow('changeLJSingleType', amber_prmtop,
                                          widget_list, var_list, description)
    cmd_window.wait_window()
    # See if we got any variables back
    vars_found = True in [bool(v.get()) for v in var_list]
    if not vars_found: return
    # addljtype expects any _non_specified variables to be None
    var_list = [v.get() for v in var_list]
    for i, v in enumerate(var_list):
        if not v: var_list[i] = None
    # If we did, store them and pass it to the class
    try:
        action = actions.changeLJSingleType(amber_prmtop,
                                                  ArgumentList(var_list))
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))
        return
    messages.write('%s\n' % action)
    action.execute()

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def addcoarsegrain(root, amber_prmtop, messages):
    """ Adds coarse graining to topology file via Lula's algo """
    # This implementation doesn't exist anywhere yet, so disable it
    showerror('Warning', 'This functionality is not implemented in Amber yet!')
    return

    # We need a file
    fname = file_chooser('Coarse Grain Parameter', 
                         [('Coarse Grain Parameters', '*.cgparm'),
                          ('All Files', '*')]
    )
   
    if not fname: return

    try:
        action = actions.addCoarseGrain(amber_prmtop, ArgumentList(fname))
        messages.write('%s\n' % action)
        action.execute()
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))
        return
   
#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def definesolvent(root, amber_prmtop, messages):
    """ Allows you to define what you consider to be solvent molecules """
    # We need a molecule #
    widget_list = [('Entry', 'List of solvent residue names')]
    var_list = [StringVar()]
    description =('Tell ParmEd that these residue names should be considered\n'
                  'to be solvent residues. This should be a comma-delimited '
                  'list')
    cmd_window = _guiwidgets.ActionWindow('defineSolvent', amber_prmtop,
                                          widget_list, var_list, description)
    cmd_window.wait_window()
    # See if we got any variables back
    var = var_list[0].get()
    if not var: return
    # addljtype expects any _non_specified variables to be None
    try:
        action = actions.defineSolvent(amber_prmtop, ArgumentList(var))
        messages.write('%s\n' % action)
        action.execute()
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))
        return

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def printbonds(root, amber_prmtop, messages):
    """ Prints bonds containing atoms from a given mask """
    widget_list = [('MaskEntry', 'Atoms 1 or Atoms to analyze for bonds'),
                   ('MaskEntry', 'Atoms 2 (optional)')]
    # We need 5 string variables, then get the description. 
    var_list = [StringVar(), StringVar()]
    description='Prints all bonds containing at least 1 atom in the given mask'
    # Create the window, open it, then wait for it to close
    cmd_window = _guiwidgets.ActionWindow('printBonds', amber_prmtop,
                                          widget_list, var_list, description)
    cmd_window.wait_window()
    # See if we got any variables back
    var = ' '.join([x.get() for x in var_list]).strip()
    if not var: return
    try:
        action = actions.printBonds(amber_prmtop, ArgumentList(var))
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))
        return
    messages.write('Printed BONDs for %s\n' % var)
    # Now make the text window
    window = Toplevel(root)
    window.resizable(True, True)
    window.title('BOND list in %s' % amber_prmtop.parm)
    text = _guiwidgets.ExitingScrollText(window, None, spacing3=5, padx=5,
                                         pady=5, width=65, height=20)
    text.pack(fill=BOTH, expand=1)
    text.write(action)
    window.wait_window()

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def printangles(root, amber_prmtop, messages):
    """ Prints angles containing atoms from a given mask """
    widget_list = [('MaskEntry', 'Atoms 1 or Atoms to analyze for angles'),
                   ('MaskEntry', 'Atoms 2 (optional)'),
                   ('MaskEntry', 'Atoms 3 (optional)')]
    var_list = [StringVar(), StringVar(), StringVar()]
    description='Prints all angles containing at least 1 atom in the given mask'
    # Create the window, open it, then wait for it to close
    cmd_window = _guiwidgets.ActionWindow('printAngles', amber_prmtop,
                                        widget_list, var_list, description)
    cmd_window.wait_window()
    # See if we got any variables back
    var = ' '.join([x.get() for x in var_list]).strip()
    if not var: return
    try:
        action = actions.printAngles(amber_prmtop, ArgumentList(var))
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))
        return
    messages.write('Printed ANGLEs for %s\n' % var)
    # Now make the text window
    window = Toplevel(root)
    window.resizable(True, True)
    window.title('ANGLE list in %s' % amber_prmtop.parm)
    text = _guiwidgets.ExitingScrollText(window, None, spacing3=5, padx=5,
                                          pady=5, width=90, height=20)
    text.pack(fill=BOTH, expand=1)
    text.write(action)
    window.wait_window()

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def printdihedrals(root, amber_prmtop, messages):
    """ Prints dihedrals containing atoms from a given mask """
    widget_list = [('MaskEntry', 'Atoms 1 or Atoms to analyze for dihedrals'),
                   ('MaskEntry', 'Atoms 2 (optional)'),
                   ('MaskEntry', 'Atoms 3 (optional)'),
                   ('MaskEntry', 'Atoms 4 (optional)')]
    var_list = [StringVar(), StringVar(), StringVar(), StringVar()]
    desc = 'Prints all dihedrals containing at least 1 atom in the given mask'
    # Create the window, open it, then wait for it to close
    cmd_window = _guiwidgets.ActionWindow('printDihedrals', amber_prmtop,
                                          widget_list, var_list, desc)
    cmd_window.wait_window()
    # See if we got any variables back
    var = ' '.join([x.get() for x in var_list]).strip()
    if not var: return
    try:
        action = actions.printDihedrals(amber_prmtop, ArgumentList(var))
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))
        return
    messages.write('Printed DIHEDRALs for %s\n' % var)
    # Now make the text window
    window = Toplevel(root)
    window.resizable(True, True)
    window.title('DIHEDRAL list in %s' % amber_prmtop.parm)
    text = _guiwidgets.ExitingScrollText(window, None, spacing3=5, padx=5,
                                         pady=5, width=140, height=20)
    text.pack(fill=BOTH, expand=1)
    text.write(action)
    window.wait_window()

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def setbond(root, amber_prmtop, messages):
    """ Sets (adds or changes) a bond in the topology file """
    # We need 2 masks, a force constant, and an equilibrium distance
    widget_list = [('MaskEntry', 'First atom in bond'),
                   ('MaskEntry', 'Second atom in bond'),
                   ('Entry', 'Force constant (kcal/mol Ang**2)'),
                   ('Entry', 'Equilibrium Distance (Ang)')]
    # We need 4 variables
    var_list = [StringVar(), StringVar(), StringVar(), StringVar()]
    description = ('Sets a bond in the topology file with the given Force '
                   'constant in kcal/mol/Ang**2\nand the given equilibrium '
                   'distance in Angstroms. Both masks must specify only a\n'
                   'single atom. If the bond exists, it will be replaced. If '
                   'it doesn\'t, it will be added.')
    # Create the window, open it, then wait for it to close
    cmd_window = _guiwidgets.ActionWindow('setBond', amber_prmtop, 
                                          widget_list, var_list, description)
    cmd_window.wait_window()
    # See if we got any variables back
    vars_found = True in [bool(v.get()) for v in var_list]
    if not vars_found: return
    # If we did, pass them through
    var_list = [v.get() for v in var_list]
    try:
        action = actions.setBond(amber_prmtop, ArgumentList(var_list))
        messages.write('%s\n' % action)
        action.execute()
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))
        return

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def setangle(root, amber_prmtop, messages):
    """ Sets (adds or changes) an angle in the topology file """
    # We need 3 masks, a force constant, and an equilibrium angle
    widget_list = [('MaskEntry', 'First atom in angle'),
                   ('MaskEntry', 'Second (middle) atom in angle'),
                   ('MaskEntry', 'Third atom in angle'),
                   ('Entry', 'Force constant (kcal/mol rad**2)'),
                   ('Entry', 'Equilibrium Angle (Degrees)')]
    # We need 5 variables
    var_list = [StringVar(), StringVar(), StringVar(), StringVar(), StringVar()]
    description = ('Sets an angle in the topology file with the given Force '
                   'constant in kcal/mol/rad**2\nand the given equilibrium '
                   'angle in Degrees. All three masks must specify only a\n'
                   'single atom. If the angle exists, it will be replaced. If '
                   'it doesn\'t, it will be added.')
    # Create the window, open it, then wait for it to close
    cmd_window = _guiwidgets.ActionWindow('setAngle', amber_prmtop, 
                                          widget_list, var_list, description)
    cmd_window.wait_window()
    # See if we got any variables back
    vars_found = True in [bool(v.get()) for v in var_list]
    if not vars_found: return
    # If we did, pass them through
    var_list = [v.get() for v in var_list]
    try:
        action = actions.setAngle(amber_prmtop, ArgumentList(var_list))
        messages.write('%s\n' % action)
        action.execute()
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))
        return

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def adddihedral(root, amber_prmtop, messages):
    """ Adds a dihedral (improper, multiterm, or normal) to the prmtop """
    # We need 4 masks, phi_k, periodicity, phase, scee/scnb, dihedral type
    widget_list = [
            ('MaskEntry', 'First (end) atom in dihedral'),
            ('MaskEntry', 'Second (middle) atom in dihedral'),
            ('MaskEntry', 'Third (middle) atom in dihedral'),
            ('MaskEntry', 'Fourth (end) atom in dihedral'),
            ('Entry', 'Phi Force constant (kcal/mol)'),
            ('Entry', 'Periodicity'),
            ('Entry', 'Phase (Degrees)'),
            ('Entry', 'EEL scaling factor'),
            ('Entry', 'VDW scaling factor'),
            ('Spinbox', 'Dihedral type', 'normal', 'improper')
    ]
    # We need 10 variables
    var_list = [StringVar() for i in range(10)]
    description = ('Adds a dihedral in the topology file with the given Phi '
                   'Force constant in kcal/mol the\ngiven phase in Degrees '
                   'and the given periodicity. All masks must specify only \n'
                   'a single atom. The default Amber values for SCEE/SCNB are '
                   '1.2 and 2.0, respectively.\nSee the Amber manual for '
                   'details about normal, multiterm, and improper dihedrals')
    # Create the window, open it, then wait for it to close
    cmd_window = _guiwidgets.ActionWindow('setDihedral', amber_prmtop, 
                                          widget_list, var_list, description)
    cmd_window.wait_window()
    # See if we got any variables back
    vars_found = True in [bool(v.get()) for v in var_list]
    if not vars_found: return
    # If we did, pass them through
    var_list = [v.get() for v in var_list]
    # Fill scee/scnb in with default values
    if var_list[7] is None: var_list[7] = '1.2'
    if var_list[8] is None: var_list[8] = '2.0'
    # The last argument is a keyword, so append that, then swap the last 2 args
    var_list.insert(9, 'type')
    try:
        action = actions.addDihedral(amber_prmtop, ArgumentList(var_list))
        messages.write('%s\n' % action)
        action.execute()
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))
        return
    
#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def addatomicnumber(root, amber_prmtop, messages):
    """ Asks the user if they want to add ATOMIC_NUMBER to the prmtop """
    response = askyesno('addAtomicNumber',
                        'Do you want to add the ATOMIC_NUMBER section to %s?' % 
                        amber_prmtop.parm)
    if response:
        action = actions.addAtomicNumber(amber_prmtop, ArgumentList(''))
        action.execute()
        messages.write('%s\n' % action)

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def deletedihedral(root, amber_prmtop, messages):
    """ Deletes a dihedral between 4 given atoms """
    # We need 4 masks
    widget_list = [('MaskEntry', 'First (end) atom in dihedral'),
                   ('MaskEntry', 'Second (middle) atom in dihedral'),
                   ('MaskEntry', 'Third (middle) atom in dihedral'),
                   ('MaskEntry', 'Fourth (end) atom in dihedral')]
    # We need 4 variables
    var_list = [StringVar() for i in range(4)]
    description = ('Deletes dihedrals between the atoms specified in mask1, ' +
                   'mask2, mask3, and mask4.\nIt will try to match dihedrals ' +
                   'only in the order given or reverse order. Dihedrals are\n' +
                   'specified by atom N in mask1, mask2, mask3, and mask4,' +
                   'where N is the selected\natom in that mask.')
    # Create the window, open it, then wait for it to close
    cmd_window = _guiwidgets.ActionWindow('deleteDihedral', amber_prmtop, 
                                          widget_list, var_list, description)
    cmd_window.wait_window()
    # See if we got any variables back
    vars_found = True in [bool(v.get()) for v in var_list]
    if not vars_found: return
    # If we did, pass them through
    var_list = [v.get() for v in var_list]
    try:
        action = actions.deleteDihedral(amber_prmtop,
                                              ArgumentList(var_list))
        messages.write('%s\n' % action)
        action.execute()
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))
        return

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def printljmatrix(root, amber_prmtop, messages):
    """ Prints A/B matrix coefficients """
    # We need 1 mask
    widget_list = [('MaskEntry', 'Atoms To Find LJ Interactions')]
    # We need 1 variable
    var_list = [StringVar()]
    description = ('Prints all A- and B-Coefficient elements between the atom '
                   'types specified in <mask> with every other atom type')
    # Create the window, open it, and wait for it to close
    cmd_window = _guiwidgets.ActionWindow('printLJMatrix', amber_prmtop, 
                                          widget_list, var_list, description)
    cmd_window.wait_window()
    # Bail out if we cancelled or something
    var = var_list[0].get()
    if not var: return
    try:
        action = actions.printLJMatrix(amber_prmtop, ArgumentList(var))
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))
        return

    # Create our Info window
    window = Toplevel(root)
    window.resizable(True, True)
    window.title('LJ Type List')
    text = _guiwidgets.ExitingScrollText(window, None, spacing3=2, padx=5,
                                         pady=5, width=45, height=30)
    text.pack(fill=BOTH, expand=1)
    text.write(action)
    messages.write('Printed LJ matrix for types [%s]\n' % var)
    window.wait_window()

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def timerge(root, amber_prmtop, messages):
    """ Merges a topology file with 2 molecules into a single prmtop for TI """
    # We need coordinates, so check for that
    if amber_prmtop.parm.coords is None:
        showerror(root, 'tiMerge requires you to load coordinates first!')
        return
    # We need 2 masks, a force constant, and an equilibrium distance
    widget_list = [('MaskEntry', 'Mask Unique to Lambda = 0'),
                   ('MaskEntry', 'Mask Unique to Lambda = 1'),
                   ('MaskEntry', 'Softcore Mask in Lambda = 0'),
                   ('MaskEntry', 'Softcore Mask in Lambda = 1'),
                   ('MaskEntry', 'Softcore Mols NOT to Merge (Lambda=0)'),
                   ('MaskEntry', 'Softcore Mols NOT to Merge (Lambda=1)'),
                   ('Entry', 'Tolerance')]
    # We need 4 variables
    var_list = [StringVar() for i in range(7)]
    var_list[6].set('0.0001')
    description = 'Merges 2 molecules inside a prmtop for use with softcore TI'
    # Create the window, open it, then wait for it to close
    cmd_window = _guiwidgets.ActionWindow('tiMerge', amber_prmtop, 
                                          widget_list, var_list, description)
    cmd_window.wait_window()
    # See if we got any variables back
    vars_found = True in [bool(v.get()) for v in var_list]
    if not vars_found: return
    # If we did, pass them through
    var_list = [v.get() for v in var_list]
    if not var_list[6].strip():
        var_list[6] = '0.0001'
    var_list.insert(6, 'tol')
    actions.tiMerge.output = messages
    try:
        action = actions.tiMerge(amber_prmtop, ArgumentList(var_list))
        messages.write('%s\n' % action)
        action.execute()
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))
        return

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def addpdb(root, amber_prmtop, messages):
    """ Adds PDB information to a topology file """

    # Get the name of the file
    type_list = [('PDB', '*.pdb'), ('All Files', '*')]
    fname = file_chooser('PDB File', type_list)
    if not fname: return

    widget_list = [('Checkbutton', 'Require all residue names be consistent.'),
                   ('Checkbutton', 'Print element names in %FLAG ELEMENT'),
                   ('Checkbutton', 'Print all insertion codes, even if all are '
                                   'blank.')]
   
    var_list = [StringVar(value='no') for i in range(3)]
    description = ('Adds information from the PDB file (e.g., CHAIN IDs) to '
                   'the topology file')
    # Create the window, open it, then wait for it to close
    cmd_window = _guiwidgets.ActionWindow('addPDB', amber_prmtop, widget_list,
                                          var_list, description)
    cmd_window.wait_window()
    # See if we got any variables back
    vars_found = True in [bool(v.get()) for v in var_list]
    if not vars_found: return
    newvars = [fname]
    # If we did, pass them through
    if var_list[0] == 'yes': newvars.append('strict')
    if var_list[1] == 'yes': newvars.append('elem')
    if var_list[1] == 'yes': newvars.append('allicodes')
    try:
        action = actions.addPDB(amber_prmtop, ArgumentList(newvars))
        messages.write('%s\n' % action)
        action.execute()
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))
        return

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def deletepdb(root, amber_prmtop, messages):
    """ Asks the user if they want to delete the sections added by addPDB """
    response = askyesno('deletePDB',
                        'Do you want to delete the addPDB flags from %s?' %
                        amber_prmtop.parm)
    if response:
        action = actions.deletePDB(amber_prmtop, ArgumentList(''))
        action.execute()
        messages.write('%s\n' % action)

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def lmod(root, amber_prmtop, messages):
    """ Prepares a prmtop file for use with lmod """
    response = askyesno('lmod', 'Do you want to adjust LJ A-coefficients for '
                                'use with lmod?')
    if response:
        action = actions.lmod(amber_prmtop, ArgumentList(''))
        action.execute()
        messages.write('%s\n' % action)

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def scale(root, amber_prmtop, messages):
    """ Allows us to scale every value in a particular prmtop section """
    # The spinbox is sent with the Spinbox, label, and then a list of all of the
    # values to give to it
    changeable_properties = ['Spinbox', 'Property to scale']
    for flag in amber_prmtop.parm.flag_list:
        if amber_prmtop.parm.formats[flag].type is float:
            changeable_properties.append(flag)
    widget_list = [changeable_properties, ('Entry', 'Value to scale by')]
    # We need 2 string variables, then get the description
    var_list = [StringVar(), StringVar()]
    description = 'Scales all values in a given prmtop section by a given value'
    # Create the window, open it, then wait for it to close
    cmd_window = _guiwidgets.ActionWindow('scale', amber_prmtop,
                                          widget_list, var_list, description)
    cmd_window.wait_window()
    # See if we got any variables back
    vars_found = var_list[0].get() or var_list[1].get()
    if not vars_found: return
    # If we did, store them and pass it to the class
    var_list = [v.get() for v in var_list]
    try:
        action = actions.scale(amber_prmtop, ArgumentList(var_list))
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))
        return
    action.execute()
    messages.write('%s\n' % action)

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def summary(root, amber_prmtop, messages):
    """ Prints a summary of the topology file to the messages """
    summary = actions.summary(amber_prmtop, ArgumentList(''))
    messages.write('%s\n' % summary)

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def hmassrepartition(root, amber_prmtop, messages):
    """
    Repartitions mass by making H-atoms heavier and attached heteroatoms
    lighter by the same amount to preserve the same total mass
    """
    # The spinbox is sent with the Spinbox, label, and then a list of all of the
    # values to give to it
    widget_list = [('Entry', 'New hydrogen mass (daltons)'),
                   ('Checkbutton', 'Repartition waters?')]
    # We need 2 string variables, then get the description
    var_list = [StringVar(value='3.024'), StringVar(value='no')]
    description = ('Repartitions atomic masses to keep total mass the same '
                   'while increasing H-atom masses to allow using a larger '
                   'time step.')
    # Create the window, open it, then wait for it to close
    cmd_window = _guiwidgets.ActionWindow('HMassChange', amber_prmtop,
                                          widget_list, var_list, description)
    cmd_window.wait_window()
    # See if we got any variables back
    vars_found = var_list[0].get() or var_list[1].get()
    if not vars_found: return
    # If we did, store them and pass it to the class
    var_list = [v.get() for v in var_list]
    # Convert the check button into the correct keyword
    if var_list[1] == 'yes':
        var_list[1] = 'dowater'
    else:
        var_list[1] = ''
    try:
        action = actions.HMassRepartition(amber_prmtop,
                                                ArgumentList(var_list))
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))
        return
    action.execute()
    messages.write('%s\n' % action)

#~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~+~

def add12_6_4(root, amber_prmtop, messages):
    """
    Adds 12-6-4 potential energy terms
    """
    # The variables we need for changeljpair
    widget_list = [('MaskEntry', 'Divalent ion mask'),
                   ('FileSelector', 'C4 Parameter File'),
                   ('Entry', 'Water model (instead of C4 Params)'),
                   ('FileSelector', 'Pol. Param File'),
                   ('Entry', 'tunfactor')]
    # Variable list -- we need 2 masks and 2 floats
    var_list = [StringVar(), StringVar(), StringVar(), StringVar(), StringVar()]
    var_list[1].set('Pick C4 File')
    var_list[3].set('Pick Pol. File')
    # description
    description = ('Add the r^-4 Lennard-Jones parameter for the 12-6-4 term\n'
                   'used typically for multi-valent ion parameters')
    cmd_window = _guiwidgets.ActionWindow('add12_6_4', amber_prmtop,
                                          widget_list, var_list, description)
    cmd_window.wait_window()
    # Make sure we didn't cancel (or just press OK with no input), or just leave
    vars_exist = [bool(v.get()) for v in var_list]
    if var_list[1].get() == 'Pick C4 File': vars_exist[1] = False
    if var_list[3].get() == 'Pick Pol. File': vars_exist[3] = False
    if vars_exist[1] and vars_exist[2]:
        showerror('Cannot select both C4 parameter file AND water model')
        return
    vars_exist = True in vars_exist
    if not vars_exist: return
    # Now that we did something, do it
    var_list = [v.get() for v in var_list]
    if var_list[1] == 'Pick C4 File': var_list[1] = ''
    if var_list[3] == 'Pick Pol. File': var_list[3] = ''
    args = [var_list[0]]
    if var_list[1]: args.extend(['c4file', var_list[1]])
    if var_list[2]: args.extend(['watermodel', var_list[2]])
    if var_list[3]: args.extend(['polfile', var_list[3]])
    if var_list[4]: args.extend(['tunfactor', var_list[4]])
    try:
        action = actions.add12_6_4(amber_prmtop, ArgumentList(args))
        messages.write('%s\n' % action)
        action.execute()
    except Exception as err:
        showerror('Unexpected Error!', '%s: %s' % (type(err).__name__, err))



"""
This is a pure-python module for reading and writing ASCII Amber structure
files, like trajectories and restarts.  Because it is pure-python and
formatted-ASCII, parsing and writing these files are expected to be slow. Binary
alternatives (like DCD and NetCDF, provided in netcdffiles.py) are strongly
encouraged, but these are provided for more complete compatibility and for
instances where the prequisites may not be installed.
"""
from __future__ import division, print_function, absolute_import

from math import ceil
import numpy as np
from parmed.formats.registry import FileFormatType
from parmed.utils.io import genopen
from parmed.utils.six import add_metaclass
from parmed.utils.six.moves import range
from parmed.utils.six import string_types
from parmed.structure import Structure
from parmed.topologyobjects import Atom
import parmed.unit as u
from parmed.vec3 import Vec3
import warnings as _warnings

VELSCALE = 20.455
ONEVELSCALE = 1 / VELSCALE

class _FileEOF(Exception):
    """ For control flow """

@add_metaclass(FileFormatType)
class _AmberAsciiCoordinateFile(object):
    """
    Abstract base class for interacting with ASCII coordinate files.
    Opens a new ASCII coordinate file and either parses it (loading
    everything into memory) or sets it up for writing.

    Parameters
    ----------
    fname : str
        File name to open
    natom : int
        Number of atoms in the system
    hasbox : bool
        Does the system have PBCs?
    mode : str={'r', 'w'}
        Whether to open this file for 'r'eading or 'w'riting
    title : str, optional
        Title to write to a new trajectory (when mode='w')
    """

    DEFAULT_TITLE = None
    CRDS_PER_LINE = None

    def __init__(self, fname, natom, hasbox, mode='r', title=None):
        if mode == 'r':
            self._status = 'old'
        elif mode == 'w':
            self._status = 'new'
            # We need to have some way to know whether we need to write the
            # coordinates or the box for this particular frame. Each frame must
            # be written as coordinates first, then box.
            self._writebox = False
        else:
            raise ValueError("%s mode must be 'r' or 'w'" % type(self).__name__)
        if isinstance(fname, string_types):
            self._file = genopen(fname, mode)
            self._own_handle = True
        elif hasattr(fname, 'read'):
            self._file = fname
            self._own_handle = False
        else:
            raise TypeError("Unsupported type for %s" % fname)

        self.natom = natom
        self.hasbox = hasbox
        if self.CRDS_PER_LINE is None:
            raise NotImplementedError('This object must be subclassed')
        self._full_lines_per_frame = self.natom * 3 // self.CRDS_PER_LINE
        self._nextras = self.natom * 3 - (self._full_lines_per_frame *
                                          self.CRDS_PER_LINE)
        if self._own_handle:
            self.closed = False
        if self._status == 'old':
            self._parse()
        elif self._status == 'new':
            if title is None:
                self._file.write('%s\n' % self.DEFAULT_TITLE)
            else:
                self._file.write(title.rstrip() + '\n')

    def _parse(self):
        """ Handles actual file parsing """
        raise NotImplementedError('virtual method not overwritten')

    @property
    def positions(self):
        return [Vec3(*xyz) for xyz in self._coordinates[0]] * u.angstroms

    def close(self):
        """ Close the open file handler """
        if self._own_handle:
            self.closed or self._file.close()
            self.closed = True

    def __del__(self):
        """ Make sure the open file handler is closed """
        try:
            self.closed or self._file.close()
        except AttributeError:
            pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class AmberAsciiRestart(_AmberAsciiCoordinateFile):
    """
    Parser for the Amber ASCII inpcrd/restart file format

    Parameters
    ----------
        fname : str
            File name to open
        mode : str={'r', 'w'}
            Whether to open this file for 'r'eading or 'w'riting
        natom : int, optional
            Number of atoms in the system (necessary when mode='w')
        hasbox : bool, optional
            Does the system have PBCs? Necessary when mode='w'
        title : str, optional
            Title to write to a new trajectory (when mode='w')
        time : float, optional
            The time to write to the restart file in ps. Default is 0.
    """
    @staticmethod
    def id_format(filename):
        """ Identifies the file type as an Amber restart/inpcrd file

        Parameters
        ----------
        filename : str
            Name of the file to check format for

        Returns
        -------
        is_fmt : bool
            True if it is an Amber restart/inpcrd file. False otherwise
        """
        if isinstance(filename, string_types):
            f = genopen(filename, 'r')
            lines = [f.readline() for i in range(5)]
            f.close()
        elif (hasattr(filename, 'readline') and hasattr(filename, 'seek')
              and hasattr(filename, 'tell')):
            cur = filename.tell()
            lines = [filename.readline() for i in range(5)]
            filename.seek(cur)
        # Look for natom
        words = lines[1].split()
        if len(words) > 2 or len(words) < 1:
            return False
        try:
            natom = int(words[0])
            float(words[1])
        except ValueError:
            return False
        except IndexError:
            pass
        # Next 3 lines, make sure we have %12.7f format. This only works if we
        # have at least 6 atoms. Any fewer than that means the restart file is
        # shorter than that.
        try:
            if natom <= 0:
                return False
            i = 0
            for line in lines[2:]:
                i += 1
                if i > natom: break
                for j in range(3):
                    j12 = j * 12
                    if line[j12+4] != '.': return False
                    float(line[j12:j12+12])
                    if line[j12+11] not in '0123456789':
                        return False
                i += 1
                if i > natom: break
                for j in range(3):
                    j12 = j * 12 + 36
                    if line[j12+4] != '.': return False
                    float(line[j12:j12+12])
                    if line[j12+11] not in '0123456789':
                        return False
        except (IndexError, ValueError):
            return False

        # Must be a restart...
        return True

    CRDS_PER_LINE = 6
    DEFAULT_TITLE = 'restart created by ParmEd'

    def __init__(self, fname, mode='r', natom=0, hasbox=None, title=None,
                 time=0.0):
        """
        For restart files, natom and hasbox are determined automatically for
        mode='r', and can be determined at write-time when the coordinates are
        set.
        """
        self._coords_written = False
        self._cell_lengths_written = False
        self._cell_angles_written = False
        self._vels_written = False
        self.time = float(time)
        super(AmberAsciiRestart, self).__init__(fname, natom, hasbox,
                                                mode, title)

    @classmethod
    def parse(cls, filename, structure=False):
        self = cls(filename)
        if structure:
            obj = Structure()
            for _ in range(self.natom):
                # fake
                obj.add_atom(Atom(), resname='XXX', resnum=0)
            obj.box = self.box
            obj.coordinates = self.coordinates[0]
            obj._coordinates = self.coordinates
            return obj
        else:
            return self

    def _parse(self):
        """
        This method parses the data out of the ASCII restart file and creates
        self._coordinates and self._velocities as np.ndarray(natom*3) arrays

        This method is called automatically for 'old' restart files and should
        not be called by external callers
        """

        lines = self._file.readlines()
        self._file.close()
        self.title = lines[0].strip()
        self.natom = int(lines[1].strip().split()[0])
        try:
            self.time = float(lines[1].strip().split()[1])
        except IndexError:
            self.time = 0.0
        # Get rid of any trailing newlines
        while not lines[-1].strip():
            lines.pop()
        # Determine what information we have based on the number of lines
        # present
        if len(lines) == int(ceil(self.natom / 2.0) + 2):
            self.hasbox = self.hasvels = False
        elif len(lines) == int(ceil(self.natom / 2.0) + 3):
            self.hasbox = True
            self.hasvels = False
        elif len(lines) == int(2 * ceil(self.natom / 2.0) + 2):
            self.hasbox = False
            self.hasvels = True
        elif len(lines) == int(2 * ceil(self.natom / 2.0) + 3):
            self.hasbox = self.hasvels = True
        else:
            raise RuntimeError('Badly formatted restart file. Has %d lines '
                               'for %d atoms.' % (len(lines), self.natom))
        self._coordinates = np.zeros((self.natom, 3))
        if self.hasvels:
            self._velocities = np.zeros((self.natom, 3))
        if self.hasbox:
            self._cell_lengths = np.zeros(3)
            self._cell_angles = np.zeros(3)
        # Now it's time to parse. Coordinates first
        startline = 2
        endline = startline + int(ceil(self.natom / 2.0))
        idx = 0
        for i in range(startline, endline):
            line = lines[i]
            x1 = float(line[ 0:12])
            y1 = float(line[12:24])
            z1 = float(line[24:36])
            self._coordinates[idx] = [x1, y1, z1]
            idx += 1
            try:
                x2 = float(line[36:48])
                y2 = float(line[48:60])
                z2 = float(line[60:72])
            except ValueError:
                pass
            else:
                self._coordinates[idx] = [x2, y2, z2]
                idx += 1
        self._coordinates = self._coordinates.reshape((1, self.natom, 3))
        startline = endline
        # Now it's time to parse the velocities if we have them
        if self.hasvels:
            endline = startline + int(ceil(self.natom / 2.0))
            idx = 0
            for i in range(startline, endline):
                line = lines[i]
                x1 = float(line[ 0:12]) * VELSCALE
                y1 = float(line[12:24]) * VELSCALE
                z1 = float(line[24:36]) * VELSCALE
                self._velocities[idx] = [x1, y1, z1]
                idx += 1
                try:
                    x2 = float(line[36:48]) * VELSCALE
                    y2 = float(line[48:60]) * VELSCALE
                    z2 = float(line[60:72]) * VELSCALE
                except ValueError:
                    pass
                else:
                    self._velocities[idx] = [x2,y2,z2]
                    idx += 1
            startline = endline
            self._velocities = self._velocities.reshape((1, self.natom, 3))
        # Now it's time to parse the box info if we have it
        if self.hasbox:
            line = lines[startline]
            self._cell_lengths[0:3] = [float(line[0:12]),
                                       float(line[12:24]),
                                       float(line[24:36])]
            self._cell_angles[0:3] = [float(line[36:48]),
                                      float(line[48:60]),
                                      float(line[60:72])]

    @property
    def coordinates(self):
        if self._status == 'new' and not hasattr(self, '_coordinates'):
            raise RuntimeError('Coordinates not yet set')
        return self._coordinates

    @coordinates.setter
    def coordinates(self, stuff):
        if self._status == 'old':
            raise RuntimeError('Cannot set coordinates on an old restart')
        stuff = np.array(stuff, copy=False).ravel()
        if self.natom > 0 and len(stuff) != 3 * self.natom:
            raise ValueError('Got %d coordinates for %d atoms' %
                             (len(stuff), self.natom))
        if self._coords_written:
            raise RuntimeError('Coordinates have already been written.')
        # Error checking done. If we didn't already set our number of atoms,
        # set that now
        self.natom = len(stuff) // 3
        self._coordinates = stuff.reshape((-1, self.natom, 3))
        self._file.write('%5d%15.7e\n' % (self.natom, self.time))
        numwrit = 0
        fmt = '%12.7f%12.7f%12.7f'
        for i in range(self.natom):
            i3 = i * 3
            self._file.write(fmt % (stuff[i3], stuff[i3+1], stuff[i3+2]))
            numwrit += 1
            if numwrit % 2 == 0: self._file.write('\n')
        if self.natom % 2 == 1: self._file.write('\n')
        self._coords_written = True

    @property
    def velocities(self):
        if self._status == 'new' and not hasattr(self, '_velocities'):
            raise RuntimeError('Velocities not set yet')
        if not self.hasvels:
            return None
        return self._velocities

    @velocities.setter
    def velocities(self, stuff):
        if self._status == 'old':
            raise RuntimeError('Cannot set velocities on an old restart')
        stuff = np.array(stuff, copy=False).ravel()
        if not self._coords_written:
            raise RuntimeError('Coordinates must be set before velocities')
        if self._cell_lengths_written or self._cell_angles_written:
            raise RuntimeError('Velocities must be written before the box info')
        if self._vels_written:
            raise RuntimeError('Can only write velocities once')
        if len(stuff) != 3 * self.natom:
            raise ValueError('Got %d velocities for %d atoms.' %
                             (len(stuff), self.natom))
        self._velocities = stuff.reshape((-1, self.natom, 3))
        fmt = '%12.7f%12.7f%12.7f'
        numwrit = 0
        for i in range(self.natom):
            i3 = i * 3
            self._file.write(fmt % (stuff[i3  ]*ONEVELSCALE,
                                    stuff[i3+1]*ONEVELSCALE,
                                    stuff[i3+2]*ONEVELSCALE)
            )
            numwrit += 1
            if numwrit % 2 == 0: self._file.write('\n')
        if self.natom % 2 == 1: self._file.write('\n')
        self._vels_written = self.hasvels = True

    @property
    def cell_lengths(self):
        if self._status == 'new' and not hasattr(self, '_cell_lengths'):
            raise RuntimeError('Cell lengths not yet available')
        if not self.hasbox:
            return None
        return self._cell_lengths

    @property
    def cell_angles(self):
        if self._status == 'new' and not hasattr(self, '_cell_angles'):
            raise RuntimeError('Cell angles not yet available')
        if not self.hasbox:
            return None
        return self._cell_angles

    @property
    def box(self):
        """ Combined cell lengths and cell angles """
        if self._status == 'new' and not (hasattr(self, '_cell_lengths') and
                hasattr(self, '_cell_angles')):
            raise RuntimeError('Cell parameters not yet set')
        if not self.hasbox:
            return None
        box = np.zeros(6)
        lengths, angles = self.cell_lengths, self.cell_angles
        box[0:3] = [lengths[0], lengths[1], lengths[2]]
        box[3:6] = [angles[0], angles[1], angles[2]]
        return box

    @cell_lengths.setter
    def cell_lengths(self, stuff):
        if self._status == 'old':
            raise RuntimeError('Cannot set cell lengths on old restart')
        if not self._coords_written:
            raise RuntimeError('Coordinates must be written before box')
        if self._cell_lengths_written:
            raise RuntimeError('Can only write cell lengths once')
        if len(stuff) != 3:
            raise ValueError('Expected 3 numbers for cell lengths')
        self._cell_lengths = np.array(stuff, copy=False)
        self._file.write('%12.7f%12.7f%12.7f' % (stuff[0], stuff[1], stuff[2]))
        self._cell_lengths_written = True

    @cell_angles.setter
    def cell_angles(self, stuff):
        if self._status == 'old':
            raise RuntimeError('Cannot set cell angles on old restart')
        if not self._coords_written:
            raise RuntimeError('Coordinates must be written before box')
        if not self._cell_lengths_written:
            raise RuntimeError('Must write cell lengths before angles')
        if self._cell_angles_written:
            raise RuntimeError('Can only write cell angles once')
        if len(stuff) != 3:
            raise ValueError('Expected 3 numbers for cell angles')
        self._cell_angles = np.array(stuff, copy=False)
        self._file.write('%12.7f%12.7f%12.7f\n' % (stuff[0],stuff[1],stuff[2]))
        self._cell_angles_written = True

    @box.setter
    def box(self, stuff):
        """ Writes both the cell lengths and cell angles """
        self.cell_lengths = stuff[:3]
        self.cell_angles = stuff[3:]

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class AmberMdcrd(_AmberAsciiCoordinateFile):
    """
    A class to parse Amber ASCII trajectory files. This is *much* slower than
    parsing NetCDF files (or the equivalent parsing done in a compiled language
    like C or C++). For large trajectories, this may be significant.
    """
    extra_args = ('natom', 'hasbox')

    CRDS_PER_LINE = 10
    DEFAULT_TITLE = 'trajectory created by ParmEd'

    @staticmethod
    def id_format(filename):
        """ Identifies the file type as an Amber mdcrd file

        Parameters
        ----------
        filename : str
            Name of the file to check format for

        Returns
        -------
        is_fmt : bool
            True if it is an Amber mdcrd file. False otherwise
        """
        f = genopen(filename, 'r')
        lines = [f.readline() for i in range(5)]
        f.close()
        # Next 4 lines, make sure we have %8.3f format
        try:
            for i in range(4):
                i += 1
                for j in range(10):
                    j8 = j * 8
                    if lines[i][j8+4] != '.': return False
                    float(lines[i][j8:j8+8])
                    if lines[i][j8+7] not in '0123456789':
                        return False
        except (IndexError, ValueError):
            return False

        # Must be a mdcrd
        return True

    def _parse(self):
        """
        This method parses the data out of the mdcrd file and creates
        self._coordinates as a list of np.ndarray(1, natom, 3) and
        self.cell_lengths as a list of np.ndarray(3) for each frame in the
        trajectory. This method is called automatically for 'old' trajectory
        files and should not be called by external callers.
        """
        line = self._file.readline()
        self.title = line.strip()
        self._coordinates = np.ndarray(0)
        self.cell_lengths = np.ndarray(0)
        mainiter = range(0, 8*self.CRDS_PER_LINE, 8)
        extraiter = range(0, 8*self._nextras, 8)
        try:
            while line:
                frame = np.zeros(self.natom*3)
                if not line: raise _FileEOF()
                cell = np.zeros(3)
                idx = 0
                line = self._file.readline()
                if not line: raise StopIteration()
                for i in range(self._full_lines_per_frame):
                    if not line: raise _FileEOF()
                    frame[idx:idx+10] = [float(line[j:j+8]) for j in mainiter]
                    idx += 10
                    line = self._file.readline()

                if self._nextras:
                    frame[idx:idx+self._nextras] = \
                            [float(line[j:j+8]) for j in extraiter]

                if self.hasbox:
                    line = self._file.readline()
                    if not line: raise _FileEOF()
                    cell[0] = float(line[:8])
                    cell[1] = float(line[8:16])
                    cell[2] = float(line[16:24])

                self._coordinates = np.concatenate((self._coordinates, frame))
                self.cell_lengths = np.concatenate((self.cell_lengths, cell))

        except _FileEOF:
            _warnings.warn('Unexpected EOF in parsing mdcrd. natom and/or '
                           'hasbox are likely wrong', RuntimeWarning)
        except StopIteration:
            pass

        self._coordinates = self._coordinates.reshape((-1, self.natom, 3))
        self.cell_lengths = self.cell_lengths.reshape((-1, 3))
        self._file.close()

    @property
    def coordinates(self):
        return self._coordinates

    @property
    def frame(self):
        return self._coordinates.shape[0]

    @property
    def box(self):
        if not self._status == 'old':
            raise RuntimeError('Cannot access box of a new mdcrd')
        return self.cell_lengths

    def add_coordinates(self, stuff):
        """
        Prints 'stuff' (which must be either an iterable of 3*natom or have an
        attribute 'flatten' that converts it into an iterable of 3*natom) to the
        open file handler. Can only be called on a 'new' mdcrd, and adds these
        coordinates to the current end of the file.

        Parameters
        ----------
        stuff : array or iterable
            This must be an iterable of length 3*natom or a numpy array that can
            be flattened to a 3*natom-length array

        Raises
        ------
        If the coordinate file is an old one being parsed or if you are
        currently expected to provide unit cell dimensions, a RuntimeError is
        raised. If the provided coordinate data does not have length 3*natom, or
        cannot be ``flatten()``ed to create a 3*natom array, a ValueError is
        raised.
        """
        # Make sure we can write the coordinates right now
        if not self._status == 'new':
            raise RuntimeError('Cannot print frames to an old mdcrd')
        try:
            stuff = stuff.flatten()
        except AttributeError:
            pass
        if self._writebox:
            raise RuntimeError('Box information not written for last frame')
        if len(stuff) != 3*self.natom:
            raise ValueError('add_coordinates requires an array of length '
                             'natom*3')

        # If we can, write the coordinates
        i = 0
        j = -1
        for i in range(self._full_lines_per_frame):
            i10 = i * 10
            for j in range(10):
                self._file.write('%8.3f' % stuff[i10+j])
            self._file.write('\n')
        if self._nextras:
            extra = i*10+j + 1
            while extra < self.natom*3:
                self._file.write('%8.3f' % stuff[extra])
                extra += 1
            self._file.write('\n')
        # Now it's time to write the box info if necessary
        self._writebox = self.hasbox
        self._file.flush()

    def add_box(self, stuff):
        """
        Prints 'stuff' (which must be a 3-element list, array.array, tuple, or
        np.ndarray) as the box lengths for this frame

        Parameters
        ----------
        stuff : array or iterable
            This must be an iterable of length 3 with the box lengths

        Raises
        ------
        If the coordinate file is an old one being parsed or if you are
        currently expected to provide coordinates, a RuntimeError is raised.
        raised. If the provided box lengths are not length 3, a ValueError is
        raised.
        """
        # First make sure we should be writing our box now
        if not self._status == 'new':
            raise RuntimeError('Cannot print box to an old mdcrd')
        if not self._writebox:
            raise RuntimeError('Should not be writing box info right now')
        if len(stuff) != 3:
            raise ValueError('add_box requires an array of length 3')

        self._file.write('%8.3f%8.3f%8.3f\n' % (stuff[0], stuff[1], stuff[2]))
        self._writebox = False
        self._file.flush()

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++



"""
This package contains code for reading CHARMM structure files for setting up a
simulation with CHARMM; specifically PSF, PAR, RTF, and STR files

    - PAR : Parameter file (PRM) -- this contains all of the force field
            parameters (e.g., bond lengths and strengths) for all of the atom
            types

    - RTF : Residue Topology File -- this file contains the residue
            connectivity tables as well as a definition of all of the atom
            types. Also contains an internal coordinate representation of the
            residues

    - PSF : Protein Structure File -- this is the main file type in CHARMM
            simulations that defines all of the residues in a system as well as
            the atom types and connectivity between the atoms

    - STR : Stream file -- Source of additional information and CHARMM commands
            that can contain RTF and PAR information. Allows users to define
            additional parameters without 'contaminating' the original force
            field parameter files
"""

__authors__ = 'Jason Swails'
__contributors__ = 'Jason Deckman'
__all__ = ['psf', 'parameters', 'charmmcrds', 'CharmmPsfFile',
           'CharmmParameterSet', 'CharmmCrdFile', 'CharmmRstFile']

from parmed.charmm.psf import CharmmPsfFile
from parmed.charmm.parameters import CharmmParameterSet
from parmed.charmm.charmmcrds import CharmmCrdFile, CharmmRstFile



print("import: 'parmed'")
import parmed

print("import: 'parmed.amber'")
import parmed.amber

print("import: 'parmed.amber.mdin'")
import parmed.amber.mdin

print("import: 'parmed.charmm'")
import parmed.charmm

print("import: 'parmed.formats'")
import parmed.formats

print("import: 'parmed.formats.pdbx'")
import parmed.formats.pdbx

print("import: 'parmed.gromacs'")
import parmed.gromacs

print("import: 'parmed.modeller'")
import parmed.modeller

print("import: 'parmed.namd'")
import parmed.namd

print("import: 'parmed.openmm'")
import parmed.openmm

print("import: 'parmed.rdkit'")
import parmed.rdkit

print("import: 'parmed.rosetta'")
import parmed.rosetta

print("import: 'parmed.tinker'")
import parmed.tinker

print("import: 'parmed.tools'")
import parmed.tools

print("import: 'parmed.tools.gui'")
import parmed.tools.gui

print("import: 'parmed.tools.simulations'")
import parmed.tools.simulations

print("import: 'parmed.unit'")
import parmed.unit

print("import: 'parmed.utils'")
import parmed.utils

print("import: 'parmed.utils.fortranformat'")
import parmed.utils.fortranformat




"""
This module contains functionality for running simulations with OpenMM using
parameters defined in an input file for sander and pmemd.
"""
from __future__ import division

import logging
from ...amber import AmberMdcrd, AmberMask, NetCDFTraj, Rst7
from ...amber.mdin import Mdin
from ...openmm import (StateDataReporter, NetCDFReporter, MdcrdReporter, RestartReporter,
                           ProgressReporter, EnergyMinimizerReporter)
from ...utils.timer import Timer
from ... import unit as u
from ...utils.six.moves import range
from math import sqrt
import os
from ..exceptions import SimulationError, SimulationWarning, UnhandledArgumentWarning
import sys
import warnings
try:
    from simtk.openmm.vec3 import Vec3
    from simtk.openmm.app import (forcefield as ff, OBC1, OBC2, GBn, HCT, GBn2,
                                  Simulation, DCDReporter, amberprmtopfile)
    import simtk.openmm as mm
    HAS_OPENMM = True
except ImportError:
    HAS_OPENMM = False

LOGGER = logging.getLogger(__name__)

_SCRIPT_HEADER = """\
#!/usr/bin/env python
from __future__ import division, print_function

import os, sys

# Import the OpenMM modules that are necessary
from simtk.openmm.app import *
from simtk.openmm import *

# Import the Amber/OpenMM modules
from parmed.amber import AmberParm, Rst7, AmberMdcrd, AmberMask, NetCDFTraj
from parmed.openmm import (StateDataReporter, NetCDFReporter, MdcrdReporter,
        RestartReporter, ProgressReporter, EnergyMinimizerReporter)
from parmed import unit as u
from parmed.utils.six.moves import range

# Load the Amber topology file
parm = AmberParm('%s', '%s')
"""

def positional_restraints(mask, weights, refc, scriptfile=None):
    """
    Creates a positional restraint force from CustomExternalForce using the
    coordinates given in the refc restart file for the atoms specified in 'mask'
    """
    parm = mask.parm # store the parm object
    try:
        refc = refc.reshape((-1, len(parm.atoms), 3))
    except ValueError:
        raise SimulationError('Invalid shape of coordinate array')

    # Make the (very simple) force term
    if scriptfile is not None:
        wt = weights.value_in_unit(u.kilocalories_per_mole/u.angstrom**2)
        scriptfile.write("# Create the positional restraint force\n")
        scriptfile.write("refc = refc.reshape((-1, len(parm.atoms), 3))\n")
        scriptfile.write("frc = CustomExternalForce('k*((x-x0)*(x-x0)+"
                         "(y-y0)*(y-y0)+(z-z0)*(z-z0))')\n")
        scriptfile.write("frc.addGlobalParameter('k', %s*u.kilocalorie_per_mole"
                         "/u.angstroms**2)\n" % wt)
        scriptfile.write("frc.addPerParticleParameter('x0')\n")
        scriptfile.write("frc.addPerParticleParameter('y0')\n")
        scriptfile.write("frc.addPerParticleParameter('z0')\n")
        scriptfile.write("for idx in mask.Selected():\n")
        scriptfile.write("    frc.addParticle(idx, refc[0, idx, :] / 10)\n")

    frc = mm.CustomExternalForce('k*((x-x0)*(x-x0)+(y-y0)*(y-y0)+'
                                    '(z-z0)*(z-z0))')
    frc.addGlobalParameter('k', weights)
    frc.addPerParticleParameter('x0')
    frc.addPerParticleParameter('y0')
    frc.addPerParticleParameter('z0')
    for idx in mask.Selected():
        frc.addParticle(idx, refc[0, idx, :] / 10) # angstroms to nanometers

    return frc

def simulate(parm, args):
    """ Runs a simulation using OpenMM """
    global HAS_OPENMM
    if not HAS_OPENMM:
        raise SimulationError('Could not import OpenMM!')

    # Set up timers
    timer = Timer()
    timer.add_timer('system', 'OpenMM System creation')
    # Handle Amber command-line arguments
    overwrite = args.has_key('-O')
    inputfile = args.get_key_string('-i', 'mdin')
    inpcrd = args.get_key_string('-c', None)
    outputfile = args.get_key_string('-o', 'mdout')
    trajectory = args.get_key_string('-x', 'mdcrd')
    mdvel = args.get_key_string('-v', 'mdvel')
    mdfrc = args.get_key_string('-frc', 'mdfrc')
    restart = args.get_key_string('-r', 'restrt')
    refc = args.get_key_string('-ref', 'refc')
    mdinfo = args.get_key_string('-inf', 'mdinfo')
    inptrajname = args.get_key_string('-y', 'inptraj')
    args.get_key_string('-cpin', None)
    args.get_key_string('-cprestrt', None)
    args.get_key_string('-cpout', None)
    args.get_key_string('-remlog', None)
    rem = args.get_key_int('-rem', 0)
    # Some OpenMM-specific options
    plat = args.get_key_string('-platform', None)
    prec = args.get_key_string('-precision', 'mixed')
    dcd = args.has_key('dcd')
    scriptfile = args.get_key_string('script', None)
    runsim = not args.has_key('norun')
    if args.has_key('progress'):
        LOGGER.warning('progress keyword is deprecated - set log level to INFO for progress')
    # Get any unmarked arguments
    unmarked_cmds = args.unmarked()
    if len(unmarked_cmds) > 0:
        warnings.warn("Un-handled arguments: " + ' '.join(unmarked_cmds), UnhandledArgumentWarning)

    # Open up the script file and write the header if requested
    if scriptfile is not None:
        scriptfile = open(scriptfile, 'w')
        if inpcrd is not None:
            scriptfile.write(_SCRIPT_HEADER % (parm, inpcrd))
        else:
            scriptfile.write(_SCRIPT_HEADER % (parm, parm.crdname))

    # Parse the input file
    mdin = Mdin()
    mdin.read(inputfile)

    # We run MD if imin == 0
    runmd = mdin.cntrl_nml['imin'] == 0

    # Process out all of the illegal options that are not supported here
    if rem:
        raise SimulationError('Replica exchange simulations are not supported.')
    if mdin.cntrl_nml['icnstph']:
        raise SimulationError('Constant pH simulations are not supported.')
    if mdin.cntrl_nml['ifqnt']:
        raise SimulationError('QM/MM simulations are not supported.')
    if mdin.cntrl_nml['ievb']:
        raise SimulationError('Empirical Valence Bond simulations not supported.')
    if mdin.cntrl_nml['ipb']:
        raise SimulationError('PB simulations not supported')
    if mdin.cntrl_nml['irism']:
        raise SimulationError('3D-RISM simulations are not supported.')
    if mdin.cntrl_nml['ifcr']:
        raise SimulationError('Charge relocation is not supported.')
    if mdin.cntrl_nml['icfe']:
        raise SimulationError('Thermodynamic integration is not supported.')
    if mdin.cntrl_nml['idecomp']:
        raise SimulationError('Residue decomposition is not supported.')
    if mdin.cntrl_nml['isgld']:
        raise SimulationError('Self-guided langevin dynamics is not supported.')
    if mdin.cntrl_nml['itgtmd']:
        raise SimulationError('Targeted MD is not supported.')
    if mdin.cntrl_nml['ineb']:
        raise SimulationError('Nudged elastic band is not supported.')
    if mdin.cntrl_nml['iamd']:
        raise SimulationError('Accelerated MD is not supported.')
    if mdin.cntrl_nml['ipimd']:
        raise SimulationError('Path-integral MD is not supported.')
    if mdin.cntrl_nml['ilscivr']:
        raise SimulationError('Linearized semi-classical simulations are not supported.')
    if parm.ptr('ifbox') > 0 and mdin.cntrl_nml['igb'] > 0:
        raise SimulationError('Cannot treat periodic system with GB')
    if mdin.cntrl_nml['nmropt']:
        raise SimulationError('NMR restraints are not supported.')
    if runmd and mdin.cntrl_nml['gamma_ln'] == 0 and mdin.cntrl_nml['ntt'] == 3:
        raise SimulationError('Langevin integrator needs a non-zero friction coefficient')
    if mdin.cntrl_nml['ioutfm'] not in (0, 1):
        raise SimulationError('ioutfm must be 0 or 1')
    # Catch some illegal trajectory printing options
    if mdin.cntrl_nml['ioutfm'] == 0 and dcd:
        warnings.warn('Forcefully setting ioutfm=1 to print DCD-formatted trajectories.',
                      SimulationWarning)
        mdin.change('cntrl', 'ioutfm', 1)
    if mdin.cntrl_nml['ntwv'] < 0 or mdin.cntrl_nml['ntwf'] < 0:
        if mdin.cntrl_nml['ioutfm'] != 1 or dcd:
            raise SimulationError('Only NetCDF trajectories can have velocities'
                                  ' and/or forces printed to the same file.')
    if (mdin.cntrl_nml['ntwv'] != 0 or mdin.cntrl_nml['ntwf'] != 0) and dcd:
        raise SimulationError('cannot print velocities or forces to DCD files.')
    if abs(mdin.cntrl_nml['ntwr']) < 500:
        warnings.warn('ntwr is set quite low (%d). Consider much larger values '
                      'to improve computational performance.' %
                      mdin.cntrl_nml['ntwr'], SimulationWarning)

    # Make sure no files will be overwritten unless overwrite is enabled
    if not overwrite:
        ERROR_MESSAGE = '%s exists. Use -O to overwrite.'
        if os.path.exists(outputfile):
            raise SimulationError(ERROR_MESSAGE % outputfile)
        if os.path.exists(mdinfo):
            raise SimulationError(ERROR_MESSAGE % mdinfo)
        if os.path.exists(restart) and mdin.cntrl_nml['imin'] != 5:
            raise SimulationError(ERROR_MESSAGE % restart)
        if os.path.exists(trajectory) and runmd and mdin.cntrl_nml['ntwx'] != 0:
            raise SimulationError(ERROR_MESSAGE % trajectory)
        if os.path.exists(mdvel) and runmd and mdin.cntrl_nml['ntwv'] > 0:
            raise SimulationError(ERROR_MESSAGE % mdvel)
        if os.path.exists(mdfrc) and runmd and mdin.cntrl_nml['ntwf'] > 0:
            raise SimulationError(ERROR_MESSAGE % mdfrc)

    # Set some dependent defaults
    if mdin.cntrl_nml['ntb'] == -1:
        if mdin.cntrl_nml['igb'] == 0:
            if mdin.cntrl_nml['ntp'] > 0:
                mdin.change('cntrl', 'ntb', 2)
            else:
                mdin.change('cntrl', 'ntb', 1)
        else:
            mdin.change('cntrl', 'ntb', 0)
    if not mdin.cntrl_nml['cut']:
        if mdin.cntrl_nml['ntb'] > 0:
            mdin.cntrl_nml['cut'] = 8.0
        else:
            mdin.cntrl_nml['cut'] = 1000.0

    # Determine our cutoff and electrostatic method
    gbmeth, kappa = None, 0.0
    if mdin.cntrl_nml['ntb'] == 0:
        # Interpret cutoffs greater than 500 Angstroms as infinite
        if mdin.cntrl_nml['cut'] >= 500:
            nbmeth = ff.NoCutoff
        else:
            nbmeth = ff.CutoffNonPeriodic
        if mdin.cntrl_nml['igb'] == 1:
            gbmeth = HCT
        elif mdin.cntrl_nml['igb'] == 2:
            gbmeth = OBC1
        elif mdin.cntrl_nml['igb'] == 5:
            gbmeth = OBC2
        elif mdin.cntrl_nml['igb'] == 7:
            gbmeth = GBn
        elif mdin.cntrl_nml['igb'] == 8:
            gbmeth = GBn2
        # Inverse Debye length assuming dielectric of 78.5 and T=298.15K
        # (multiplied by 0.73 to account for lack of ion-ion exclusions)
        kappa = 0.73 * sqrt(mdin.cntrl_nml['saltcon'] * 0.10806)
    else:
        # Periodic
        if mdin.ewald_nml['ew_type'] == 0:
            nbmeth = ff.PME
        elif mdin.ewald_nml['ew_type'] == 1:
            nbmeth = ff.Ewald
        if mdin.ewald_nml['eedmeth'] == 4:
            if mdin.cntrl_nml['cut'] > 500:
                nbmeth = ff.NoCutoff
            else:
                nbmeth = ff.CutoffPeriodic
        elif mdin.ewald_nml['eedmeth'] != 1:
            warnings.warn('eedmeth must be 1 or 4. Other values are ignored.',
                          SimulationWarning)
        if mdin.ewald_nml['vdwmeth'] not in (0, 1):
            raise SimulationError('vdwmeth must be 0 or 1')

    # Determine our constraints
    if mdin.cntrl_nml['ntc'] == 1:
        constraints = None
        rw = False
    elif mdin.cntrl_nml['ntc'] == 2:
        constraints = ff.HBonds
        rw = True
    elif mdin.cntrl_nml['ntc'] == 3:
        constraints = ff.AllBonds
        rw = True
    else:
        raise SimulationError('ntc must be 1, 2, or 3')

    # Determine if these constraints are flexible (their energies are computed)
    if mdin.cntrl_nml['ntf'] == 1:
        flexconst = True
    elif mdin.cntrl_nml['ntf'] == 2:
        flexconst = False
    elif mdin.cntrl_nml['ntf'] > 0 and mdin.cntrl_nml['ntf'] < 8:
        flexconst = False
        warnings.warn('ntf > 2 just implies no constraints are flexible.')
    else:
        raise SimulationError('ntf must be between 0 and 8')

    # Create the OpenMM System object from our current options
    timer.start_timer('system')
    LOGGER.info('Setting up the OpenMM system...')
    system = parm.createSystem(nonbondedMethod=nbmeth,
                     nonbondedCutoff=mdin.cntrl_nml['cut']*u.angstrom,
                     constraints=constraints, rigidWater=rw,
                     implicitSolvent=gbmeth,
                     implicitSolventKappa=kappa*(1.0/u.angstrom),
                     soluteDielectric=mdin.cntrl_nml['intdiel'],
                     solventDielectric=mdin.cntrl_nml['extdiel'],
                     removeCMMotion=mdin.cntrl_nml['nscm']>0,
                     ewaldErrorTolerance=mdin.ewald_nml['rsum_tol'],
                     flexibleConstraints=flexconst,
                     verbose=False
    )
    # Log it if requested
    if scriptfile is not None:
        scriptfile.write("system = parm.createSystem(nonbondedMethod=%s,\n"
                "                nonbondedCutoff=%s*u.angstrom,\n"
                "                constraints=%s, rigidWater=%s,\n"
                "                implicitSolvent=%s,\n"
                "                implicitSolventKappa=%s*(1.0/u.angstrom),\n"
                "                soluteDielectric=%s,\n"
                "                solventDielectric=%s,\n"
                "                removeCMMotion=%s,\n"
                "                ewaldErrorTolerance=%s,\n"
                "                flexibleConstraints=%s,\n"
                "                verbose=False,\n)\n\n" %
                (nbmeth, mdin.cntrl_nml['cut'], constraints, rw, gbmeth, kappa,
                mdin.cntrl_nml['intdiel'], mdin.cntrl_nml['extdiel'],
                mdin.cntrl_nml['nscm']>0, mdin.ewald_nml['rsum_tol'], flexconst)
        )

    # See if we need to turn off the long-range dispersion correction
    if mdin.cntrl_nml['ntb'] > 0 and mdin.ewald_nml['vdwmeth'] == 0:
        # Disable the long-range vdW correction
        for i, frc in enumerate(system.getForces()):
            if (isinstance(frc, mm.NonbondedForce)):
                frc.setUseDispersionCorrection(False)
                if scriptfile is not None:
                    scriptfile.write('# Disable long-range vdW correction\n')
                    scriptfile.write('system.getForces()[%d].setUseDispersionCorrection(False)\n'%i)
            if isinstance(frc, mm.CustomNonbondedForce):
                frc.setUseLongRangeCorrection(False)
                if scriptfile is not None:
                    scriptfile.write('# Disable long-range vdW correction\n')
                    scriptfile.write('system.getForces()[%d].setUseLongRangeCorrection(False)\n'%i)
        if scriptfile is not None:
            scriptfile.write('\n')

    timer.stop_timer('system')

    # See if we need to add positional restraints
    if mdin.cntrl_nml['ntr'] != 0:
        timer.add_timer('restraints', 'Setting up positional restraints')
        timer.start_timer('restraints')
        LOGGER.debug('Setting up the restraints...')
        if mdin.cntrl_nml['restraint_wt'] <= 0:
            raise SimulationError('Restraint weight must be > 0 for restrained dynamics')
        if not mdin.cntrl_nml['restraintmask'].strip():
            raise SimulationError('Restrained atom selection must be made via restraintmask')
        if scriptfile is not None:
            scriptfile.write("# Set up the restrained atom selection mask\n")
            scriptfile.write("mask = AmberMask(parm, '%s')\n" % mdin.cntrl_nml['restraintmask'])
            scriptfile.write('refc = Rst7.open("%s")\n' % refc)
        mask = AmberMask(parm, mdin.cntrl_nml['restraintmask'])
        system.addForce(
            positional_restraints(mask, mdin.cntrl_nml['restraint_wt'] *
                                  u.kilocalorie_per_mole/u.angstrom/u.angstrom,
                                  Rst7.open(refc), scriptfile=scriptfile)
        )
        if scriptfile is not None:
            scriptfile.write('system.addForce(frc)\n\n')
        timer.stop_timer('restraints')

    # See if we need to add a barostat
    if (mdin.cntrl_nml['ntb'] == 2 or mdin.cntrl_nml['ntp'] > 0) and runmd:
        if mdin.cntrl_nml['ntt'] == 0:
            raise SimulationError('constant pressure requires a thermostat')
        if mdin.cntrl_nml['barostat'] == 1:
            raise SimulationError('Berendsen barostat is not implemented in OpenMM. Use the MC '
                                  'barostat instead (barostat=2)')
        if mdin.cntrl_nml['ntp'] == 1:
            # Isotropic scaling
            barostat = mm.MonteCarloBarostat(mdin.cntrl_nml['pres0']*u.bar,
                                             mdin.cntrl_nml['temp0']*u.kelvin,
                                             mdin.cntrl_nml['mcbarint'],
                                            )
            if scriptfile is not None:
                scriptfile.write('barostat = MonteCarloBarostat(%s*u.bar,\n                    '
                                 '%s*u.kelvin, %s\n)\n' % (mdin.cntrl_nml['pres0'],
                                 mdin.cntrl_nml['temp0'], mdin.cntrl_nml['mcbarint'])
                                )
        elif mdin.cntrl_nml['ntp'] == 2:
            # Anisotropic scaling
            barostat = mm.MonteCarloAnisotropicBarostat(mdin.cntrl_nml['pres0']*u.bar,
                                                        mdin.cntrl_nml['temp0']*u.kelvin, True,
                                                        True, True, mdin.cntrl_nml['mcbarint'],
                                                       )
            if scriptfile is not None:
                scriptfile.write('barostat = MonteCarloAnisotropicBarostat(%s*u.bar,\n  '
                                 '               %s*u.kelvin, True, True, True, %s\n)\n' %
                                 (mdin.cntrl_nml['pres0'], mdin.cntrl_nml['temp0'],
                                  mdin.cntrl_nml['mcbarint'])
                                )
        else:
            raise SimulationError('ntp must be 1 or 2 for constant pressure')
        if scriptfile is not None:
            scriptfile.write('system.addForce(barostat)\n\n')
        system.addForce(barostat)

    # Set the integrator (we need this to set constraint parameters even in
    # minimization if someone requested SHAKEn minimization). Then set the
    # tolerance and get the simulation platform
    if mdin.cntrl_nml['ntt'] == 3:
        integrator = mm.LangevinIntegrator(mdin.cntrl_nml['temp0']*u.kelvin,
                                           mdin.cntrl_nml['gamma_ln']/u.picosecond,
                                           mdin.cntrl_nml['dt']*u.picosecond
                                          )
        if scriptfile is not None:
            scriptfile.write('integrator = LangevinIntegrator(%s*u.kelvin, '
                             '%s/u.picosecond,\n                    %s*u.picosecond\n)\n' %
                             (mdin.cntrl_nml['temp0'], mdin.cntrl_nml['gamma_ln'],
                              mdin.cntrl_nml['dt'])
                            )
    else:
        if mdin.cntrl_nml['ntt'] == 1:
            raise SimulationError('ntt must be 2 or 3 for OpenMM (Andersen '
                                  'thermostat or Langevin dynamics')
        elif mdin.cntrl_nml['ntt'] == 2:
            if scriptfile is not None:
                scriptfile.write('system.addForce(AndersenThermostat(%s*u.kelvin, %s\n)\n' %
                                 (mdin.cntrl_nml['temp0'], mdin.cntrl_nml['vrand']))
            system.addForce(
                mm.AndersenThermostat(mdin.cntrl_nml['temp0']*u.kelvin, mdin.cntrl_nml['vrand'])
            )
        if scriptfile is not None:
            scriptfile.write('integrator = VerletIntegrator(%s*u.picosecond)\n'
                             % mdin.cntrl_nml['dt'])
        integrator = mm.VerletIntegrator(mdin.cntrl_nml['dt']*u.picosecond)

    # Set the constraint tolerance on the integrator
    if constraints is not None:
        integrator.setConstraintTolerance(mdin.cntrl_nml['tol'])
        if scriptfile is not None:
            scriptfile.write('integrator.setConstraintTolerance(%s)\n' % mdin.cntrl_nml['tol'])

    # Define the platform if it was requested, or just take the default
    if plat is not None:
        if plat not in ('CUDA', 'Reference', 'CPU', 'OpenCL'):
            raise SimulationError('Platform must be one of CUDA, Reference, CPU, or OpenCL')
        platform = mm.Platform.getPlatformByName(plat)
        if scriptfile is not None:
            scriptfile.write('platform = Platform.getPlatformByName("%s")\n' % plat)
    else:
        if scriptfile is not None:
            scriptfile.write('platform = None\n')
        platform = None

    if prec not in ('mixed', 'double', 'single'):
        raise SimulationError('Invalid precision model [%s]. Must be single, '
                              'double, or mixed.' % prec)

    # Create the Simulation object
    timer.add_timer('simulation', 'Creating the Simulation object')
    timer.start_timer('simulation')
    LOGGER.info('Setting up the simulation...')
    if scriptfile is not None:
        scriptfile.write('simulation = Simulation(parm.topology, system, integrator, platform)\n\n')
    simulation = Simulation(parm.topology, system, integrator, platform)

    # Now set the default precision model. This needs to be done based on the
    # chosen platform. To handle the case where we chose the default platform,
    # we need to get the platform from the current context
    if platform is None:
        platform = simulation.context.getPlatform()
        if scriptfile is not None:
            scriptfile.write('platform = simulation.context.getPlatform()\n')
    if platform.getName() == 'CUDA':
        if scriptfile is not None:
            scriptfile.write('platform.setPropertyValue(simulation.context, '
                             "'CudaPrecision', '%s')\n\n" % prec)
        platform.setPropertyValue(simulation.context, 'CudaPrecision', prec)
    elif platform.getName() == 'OpenCL':
        if scriptfile is not None:
            scriptfile.write('platform.setPropertyValue(simulation.context, '
                             "'OpenCLPrecision', '%s')\n\n" % prec)
        platform.setPropertyValue(simulation.context, 'OpenCLPrecision', prec)
    elif platform.getName() == 'Reference' and prec != 'double':
        warnings.warn('The Reference platform only uses double precision', SimulationWarning)
    elif platform.getName() == 'CPU' and prec != 'single':
        warnings.warn('The CPU platform only uses single precision', SimulationWarning)

    # Set the particle positions and box vectors (if applicable) from either the
    # given restart file or from the topology file object. Also use particle
    # velocities if requested via irest=1
    LOGGER.info('Setting up initial coordinates and velocities...')
    if inpcrd is not None:
        position_container = Rst7.open(inpcrd)
        if position_container.natom != parm.ptr('natom'):
            raise SimulationError('inpcrd [%s] and prmtop mismatch in number of atoms' % inpcrd)
    else:
        position_container = parm
    if scriptfile is not None:
        scriptfile.write('# Set the positions and box vectors\n'
                         'simulation.context.setPositions(parm.positions)\n')
    simulation.context.setPositions(position_container.positions)
    if parm.ptr('ifbox') > 0:
        # Only set box vectors if box is present
        if scriptfile is not None:
            scriptfile.write('simulation.context.setPeriodicBoxVectors(*parm.box_vectors)\n\n')
        simulation.context.setPeriodicBoxVectors(*position_container.box_vectors)

    # Velocities
    if runmd and mdin.cntrl_nml['irest'] == 1 and position_container.hasvels:
        if scriptfile is not None:
            scriptfile.write('# Set velocities\n')
            scriptfile.write('simulation.context.setVelocities(parm.velocities)\n')
        simulation.context.setVelocities(position_container.velocities)
    elif runmd:
        if scriptfile is not None:
            scriptfile.write('# Set velocities\n')
            scriptfile.write('simulation.context.setVelocitiesToTemperature(\n'
                             '           %s*u.kelvin)\n' % mdin.cntrl_nml['tempi'])
        simulation.context.setVelocitiesToTemperature(mdin.cntrl_nml['tempi']*u.kelvin)
    # Add the energy reporter
    if scriptfile is not None:
        scriptfile.write('# Add the state data reporters\n')
        density = runmd and mdin.cntrl_nml['ntp'] > 0
        scriptfile.write('rep = StateDataReporter("%s", %s, volume=%s, density=%s)\n' %
                         (outputfile, mdin.cntrl_nml['ntpr'], parm.ptr('ifbox') > 0, density)
                        )
        scriptfile.write('simulation.reporters.append(rep)\n')
        scriptfile.write('rep = ProgressReporter("%s", %s, %s, volume=%s, density=%s)\n' %
                         (mdinfo, mdin.cntrl_nml['ntpr'], mdin.cntrl_nml['nstlim'],
                          parm.ptr('ifbox') > 0, density)
                        )
        scriptfile.write('simulation.reporters.append(rep)\n')
    rep = StateDataReporter(outputfile, mdin.cntrl_nml['ntpr'], volume=parm.ptr('ifbox') > 0,
                            density=density)
    simulation.reporters.append(rep)
    rep = ProgressReporter(mdinfo, mdin.cntrl_nml['ntpr'], mdin.cntrl_nml['nstlim'],
                           volume=mdin.cntrl_nml['ntb'] > 0, density=density)
    simulation.reporters.append(rep)

    timer.stop_timer('simulation')

    # Now see if we wanted minimization or not
    if mdin.cntrl_nml['imin'] > 0:
        # We're doing minimization
        if mdin.cntrl_nml['imin'] == 1:
            timer.add_timer('minimization', 'Structure minimization')
            timer.start_timer('minimization')
            LOGGER.info('Minimizing...')
            if scriptfile is not None:
                scriptfile.write('rep = EnergyMinimizerReporter("%s", volume=%s)\n' %
                                 (outputfile, parm.ptr('ifbox') > 0))
                scriptfile.write('rep.report(simulation)\n')
                scriptfile.write('# Minimize the energy\n')
                scriptfile.write('simulation.minimizeEnergy(\n'
                                 '    tolerance=%s*u.kilocalories_per_mole,\n'
                                 '    maxIterations=%s\n)\n' %
                                 (mdin.cntrl_nml['drms'], mdin.cntrl_nml['maxcyc']))
                scriptfile.write('rep.report(simulation)\n\n')
                scriptfile.write('# Write the restart file\n')
                scriptfile.write('restrt_reporter = RestartReporter("%s", 1,\n'
                                 '    False, %s, write_velocities=False\n)\n' %
                                 (restart, mdin.cntrl_nml['ntxo']==2))
                scriptfile.write('restrt_reporter.report(simulation,\n'
                                 '    simulation.context.getState(getPositions=True, '
                                 'enforcePeriodicBox=%s)\n)\n' % bool(mdin.cntrl_nml['ntb']))
            rep = EnergyMinimizerReporter(outputfile, volume=parm.ptr('ifbox') > 0)
            rep.report(simulation)
            simulation.minimizeEnergy(tolerance=mdin.cntrl_nml['drms']*u.kilocalories_per_mole,
                                      maxIterations=mdin.cntrl_nml['maxcyc'])
            rep.report(simulation)
            # Write a restart file with the new coordinates
            restrt_reporter = RestartReporter(restart, 1, parm.ptr('natom'), False,
                                              mdin.cntrl_nml['ntxo'] == 2, write_velocities=False)
            restrt_reporter.report(simulation, simulation.context.getState(getPositions=True,
                                   enforcePeriodicBox=bool(mdin.cntrl_nml['ntb'])))
            timer.stop_timer('minimization')
        elif mdin.cntrl_nml['imin'] == 5:
            timer.add_timer('minimization', 'Structure minimization')
            timer.start_timer('minimization')
            LOGGER.info('Minimizing...')
            try:
                inptraj = NetCDFTraj.open_old(inptrajname)
                nframes = inptraj.frame
                if scriptfile is not None:
                    scriptfile.write('inptraj = NetCDFTraj.open_old("%s")\n' % inptrajname)
                    scriptfile.write('nframes = inptraj.frame\n')
            except RuntimeError:
                inptraj = AmberMdcrd(inptrajname, parm.ptr('natom'), mdin.cntrl_nml['ntb'] > 0,
                                     mode='r')
                nframes = len(inptraj.coordinates())
                if scriptfile is not None:
                    scriptfile.write('inptraj = AmberMdcrd("%s", %s, %s, mode="r")\n' %
                                     (inptrajname, "parm.ptr('natom')", mdin.cntrl_nml['ntb']>0))
                    scriptfile.write('nframes = inptraj.frame\n')
            # Create a reporter to handle the minimized coordinates
            if mdin.cntrl_nml['ioutfm'] == 0:
                crd_reporter = MdcrdReporter(trajectory, 1)
                if scriptfile is not None:
                    scriptfile.write('crd_reporter = MdcrdReporter("%s", 1)\n' % trajectory)
            elif dcd:
                crd_reporter = DCDReporter(trajectory, 1)
                if scriptfile is not None:
                    scriptfile.write('crd_reporter = DCDReporter("%s", 1)\n' % trajectory)
            else:
                crd_reporter = NetCDFReporter(trajectory, 1)
                if scriptfile is not None:
                    scriptfile.write('crd_reporter = NetCDFReporter("%s", 1)\n' % trajectory)
            if scriptfile is not None:
                scriptfile.write('f = open("%s", "w", 0)\n'
                                 'rep = EnergyMinimizerReporter(f, volume=%s)\n'
                                 'for frame in range(nframes)\n'
                                 '    crds = inptraj.coordinates(frame)\n'
                                 '    simulation.context.setPositions(\n'
                                 '         tuple([Vec3(crds[3*i], crds[3*i+1], '
                                 'crds[3*i+2])\n'
                                 '           for i in range(parm.ptr("natom"))]) * '
                                 'u.angstroms\n    )\n' % outputfile)
            f = open(outputfile, 'w', 0)
            rep = EnergyMinimizerReporter(f, volume=parm.ptr('ifbox') > 0)
            for frame in range(nframes):
                crds = inptraj.coordinates(frame)
                simulation.context.setPositions(
                        tuple([Vec3(crds[3*i], crds[3*i+1], crds[3*i+2])
                        for i in range(parm.ptr('natom'))]) * u.angstroms
                )
                rep.report(simulation, frame=frame+1)
                if mdin.cntrl_nml['maxcyc'] > 1:
                    if scriptfile is not None:
                        scriptfile.write('    simulation.minimizeEnergy(\n'
                                         '          tolerance=%s*u.kilocalories_per_mole,\n'
                                         '          maxIterations=%s,\n'
                                         '    )\n'
                                         '    crd_reporter.report(simulation,\n'
                                         '        simulation.context.getState('
                                         'getPositions=True),\n    )\n'
                                         '    rep.report(simulation)\n' %
                                         (mdin.cntrl_nml['drms'], mdin.cntrl_nml['maxcyc']))
                    simulation.minimizeEnergy(
                        tolerance=mdin.cntrl_nml['drms']* u.kilocalories_per_mole,
                        maxIterations=mdin.cntrl_nml['maxcyc']
                    )
                    crd_reporter.report(simulation, simulation.context.getState(getPositions=True))
                    rep.report(simulation)
                f.write('=' * 80 + '\n')
            f.close()
            if scriptfile is not None:
                scriptfile.write('    f.write("=" * 80 + "\\n")\n')
                scriptfile.write('f.close()\n')
            timer.stop_timer('minimization')
    # Otherwise we want to do dynamics
    else:
        timer.add_timer('md', 'Molecular Dynamics')
        timer.start_timer('md')
        # Coordinate trajectory reporters
        if mdin.cntrl_nml['ntwx'] > 0:
            if mdin.cntrl_nml['ioutfm'] == 0:
                simulation.reporters.append(
                        MdcrdReporter(trajectory, mdin.cntrl_nml['ntwx'])
                )
                if scriptfile is not None:
                    scriptfile.write('simulation.reporters.append(\n'
                                     '      MdcrdReporter("%s", %s)\n)\n' %
                                     (trajectory, mdin.cntrl_nml['ntwx']))
            elif dcd:
                simulation.reporters.append(DCDReporter(trajectory, mdin.cntrl_nml['ntwx']))
                if scriptfile is not None:
                    scriptfile.write('simulation.reporters.append(\n'
                                     '      DCDReporter("%s", %s)\n)\n' %
                                     (trajectory, mdin.cntrl_nml['ntwx']))
            else:
                simulation.reporters.append(
                   NetCDFReporter(trajectory, mdin.cntrl_nml['ntwx'], crds=True,
                                  vels=mdin.cntrl_nml['ntwv'] < 0, frcs=mdin.cntrl_nml['ntwf'] < 0)
                )
                if scriptfile is not None:
                    scriptfile.write('simulation.reporters.append(\n'
                                     '      NetCDFReporter("%s", %s,\n'
                                     '          crds=True, vels=%s, frcs=%s\n'
                                     '      )\n)\n' % (trajectory, mdin.cntrl_nml['ntwx'],
                                     mdin.cntrl_nml['ntwv'] < 0, mdin.cntrl_nml['ntwf'] < 0))
        # Velocity trajectory reporters
        if mdin.cntrl_nml['ntwv'] > 0:
            if mdin.cntrl_nml['ioutfm'] == 0:
                simulation.reporters.append(
                   MdcrdReporter(mdvel, mdin.cntrl_nml['ntwv'], crds=False, vels=True, frcs=False)
                )
                if scriptfile is not None:
                    scriptfile.write('simulation.reporters.append(\n'
                                     '      MdcrdReporter("%s", %s, crds=False,\n'
                                     '         vels=True, frcs=False)\n'
                                     ')\n' % (trajectory, mdin.cntrl_nml['ntwv']))
            elif dcd:
                # Should never make it here due to earlier checks; catch anyway
                raise SimulationError('Cannot write velocities to DCD files.')
            else:
                # Add forces too only if ntwf < 0 AND no coordinate traj is
                # being printed (otherwise forces will be printed to the
                # coordinate file
                _frcs = mdin.cntrl_nml['ntwf'] < 0 and mdin.cntrl_nml['ntwx'] == 0
                simulation.reporters.append(
                    NetCDFReporter(mdvel, mdin.cntrl_nml['ntwv'], crds=False, vels=True, frcs=_frcs)
                )
                if scriptfile is not None:
                    scriptfile.write('simulation.reporters.append(\n'
                                     '      NetCDFReporter("%s", %s,\n'
                                     '          crds=False, vels=True, frcs=%s\n'
                                     '      )\n)\n' % (mdvel, mdin.cntrl_nml['ntwv'],
                                      mdin.cntrl_nml['ntwf'] < 0 and mdin.cntrl_nml['ntwx'] == 0))
        # Force trajectory reporters
        if mdin.cntrl_nml['ntwf'] > 0:
            if mdin.cntrl_nml['ioutfm'] == 0:
                simulation.reporters.append(
                    MdcrdReporter(mdfrc, mdin.cntrl_nml['ntwf'], crds=False, vels=False, frcs=True)
                )
                if scriptfile is not None:
                    scriptfile.write('simulation.reporters.append(\n'
                                     '      MdcrdReporter("%s", %s, crds=False,\n'
                                     '         vels=False, frcs=True)\n'
                                     ')\n' % (trajectory, mdin.cntrl_nml['ntwf']))
            elif dcd:
                # Should never make it here due to earlier checks; catch anyway
                raise SimulationError('Cannot write forces to DCD files.')
            else:
                simulation.reporters.append(
                    NetCDFReporter(mdfrc, mdin.cntrl_nml['ntwf'], crds=False, vels=False, frcs=True)
                )
                if scriptfile is not None:
                    scriptfile.write('simulation.reporters.append(\n'
                                     '      NetCDFReporter("%s", %s,\n'
                                     '          crds=False, vels=False, frcs=True\n'
                                     '      )\n)\n' % (mdfrc, mdin.cntrl_nml['ntwv']))
        # Restart file reporter
        restrt_reporter = RestartReporter(restart, abs(mdin.cntrl_nml['ntwr']),
                                          mdin.cntrl_nml['ntwr'] < 0, mdin.cntrl_nml['ntxo'] == 2)
        if scriptfile is not None:
            scriptfile.write('restrt_reporter = RestartReporter("%s", %s,\n      %s, %s)\n' %
                             (restart, abs(mdin.cntrl_nml['ntwr']), mdin.cntrl_nml['ntwr'] < 0,
                              mdin.cntrl_nml['ntxo'] == 2))
        if mdin.cntrl_nml['ntwr'] != 0:
            simulation.reporters.append(restrt_reporter)
            if scriptfile is not None:
                scriptfile.write('simulation.reporters.append(restrt_reporter)\n')

        LOGGER.info('Running MD...')
        if scriptfile is not None:
            scriptfile.write('simulation.step(%s)\n'
                             'final_state = simulation.context.getState(getPositions=True,\n'
                             '     getVelocities=True, enforcePeriodicBox=%s)\n'
                             'restrt_reporter.report(simulation, final_state)\n' %
                             (mdin.cntrl_nml['nstlim'], mdin.cntrl_nml['ntb'] > 0))
        if runsim:
            simulation.step(mdin.cntrl_nml['nstlim'])
            # Now write the final restart file, hijacking the restart file
            # reporter
            final_state = simulation.context.getState(getPositions=True, getVelocities=True,
                                                      enforcePeriodicBox=mdin.cntrl_nml['ntb'] > 0)
            restrt_reporter.report(simulation, final_state)
            timer.stop_timer('md')
            nsperday = (mdin.cntrl_nml['nstlim'] * mdin.cntrl_nml['dt'] /
                        1000.0 / (timer.timers['md'] / 3600 / 24))

    timer.done()
    LOGGER.info('Done')
    for t in timer.timer_names:
        timer.print_(t, sys.stdout)

    if mdin.cntrl_nml['imin'] == 0 and runsim:
        LOGGER.info('MD timing: %.3f ns/day' % nsperday)

    if scriptfile is not None:
        scriptfile.close()

def energy(parm, args, output=sys.stdout):
    """
    Computes a single-point energy using OpenMM and prints the result to the
    desired output destination (defaults to stdout)
    """
    import tempfile
    global HAS_OPENMM
    if not HAS_OPENMM:
        raise SimulationError('Could not import OpenMM!')

    cutoff = args.get_key_float('cutoff', None)
    igb = args.get_key_int('igb', 5)
    saltcon = args.get_key_float('saltcon', 0.0)
    do_ewald = args.has_key('Ewald')
    plat = args.get_key_string('platform', None)
    prec = args.get_key_string('precision', 'mixed')
    decomp = args.has_key('decompose')
    applayer = args.has_key('applayer')
    vdw_longrange = not args.has_key('nodisper')
    # Get any unmarked arguments
    unmarked_cmds = args.unmarked()
    if len(unmarked_cmds) > 0:
        warnings.warn("Un-handled arguments: " + ' '.join(unmarked_cmds), UnhandledArgumentWarning)

    gbmeth = None
    if parm.ptr('ifbox') == 0:
        if cutoff is None or cutoff >= 500:
            nbmeth = ff.NoCutoff
            cutoff = 1000.0
        else:
            nbmeth = ff.CutoffNonPeriodic
        if not igb in (0, 1, 2, 5, 6, 7, 8):
            raise SimulationError('Bad igb value. Must be 0, 1, 2, 5, 6, 7, or 8')
        if igb == 1:
            gbmeth = HCT
        elif igb == 2:
            gbmeth = OBC1
        elif igb == 5:
            gbmeth = OBC2
        elif igb == 7:
            gbmeth = GBn
        elif igb == 8:
            gbmeth = GBn2
    else:
        cutoff = cutoff if cutoff is not None else 8.0
        if do_ewald:
            nbmeth = ff.Ewald
        else:
            nbmeth = ff.PME

    parm_ = parm
    if applayer:
        # Write out a temporary topology file, load an amberprmtopfile, then
        # delete that file
        tmp = tempfile.mktemp(suffix='.parm7')
        try:
            parm.write_parm(tmp)
            parm_ = amberprmtopfile.AmberPrmtopFile(tmp)
            os.unlink(tmp)
        except IOError:
            raise SimulationError('Could not create temporary file for app ' # pragma: no cover
                                  'layer energy calculation.')

    # Time to create the OpenMM system
    system = parm_.createSystem(nonbondedMethod=nbmeth, nonbondedCutoff=cutoff*u.angstrom,
                                constraints=None, rigidWater=True, removeCMMotion=False,
                                implicitSolvent=gbmeth, implicitSolventSaltConc=saltcon*u.molar,
                                soluteDielectric=1.0, solventDielectric=78.5,
                                ewaldErrorTolerance=5e-5)

    # If we used the app layer, we need to assign force groups to enable energy
    # decomposition
    if applayer:
        for force in system.getForces():
            if isinstance(force, mm.HarmonicBondForce):
                force.setForceGroup(parm.BOND_FORCE_GROUP)
            elif isinstance(force, mm.HarmonicAngleForce):
                force.setForceGroup(parm.ANGLE_FORCE_GROUP)
            elif isinstance(force, mm.PeriodicTorsionForce):
                force.setForceGroup(parm.DIHEDRAL_FORCE_GROUP)
            else:
                # Treat the rest as nonbonded forces
                force.setForceGroup(parm.NONBONDED_FORCE_GROUP)
        # For periodic simulations, we need to set the box info
        if parm.ptr('ifbox') > 0:
            system.setDefaultPeriodicBoxVectors(*parm.box_vectors)

    # Now see if we want to turn on or off the dispersion correction
    for force in system.getForces():
        if isinstance(force, mm.NonbondedForce):
            force.setUseDispersionCorrection(vdw_longrange)
        if isinstance(force, mm.CustomNonbondedForce):
            force.setUseLongRangeCorrection(vdw_longrange)

    # Create a dummy integrator
    integrator = mm.VerletIntegrator(2.0)

    # Define the platform if it was requested, or just take the default
    if plat is not None:
        if plat not in ('CUDA', 'Reference', 'CPU', 'OpenCL'):
            raise SimulationError('Platform must be one of CUDA, Reference, CPU, or OpenCL')
        platform = mm.Platform.getPlatformByName(plat)
    else:
        platform = None

    if prec not in ('mixed', 'double', 'single'):
        raise SimulationError('Invalid precision model [%s]. Must be single, '
                              'double, or mixed.' % prec)

    # Create the Simulation object
    if platform is None:
        context = mm.Context(system, integrator)
    else:
        context = mm.Context(system, integrator, platform)

    # Set the particle positions
    context.setPositions(parm.positions)

    # Now set the default precision model. This needs to be done based on the
    # chosen platform. To handle the case where we chose the default platform,
    # we need to get the platform from the current context
    if platform is None:
        platform = context.getPlatform()
    if platform.getName() == 'CUDA':
        platform.setPropertyValue(context, 'CudaPrecision', prec)
    elif platform.getName() == 'OpenCL':
        platform.setPropertyValue(context, 'OpenCLPrecision', prec)
    elif platform.getName() == 'Reference' and prec != 'double':
        warnings.warn('The Reference platform only uses double precision', SimulationWarning)
    elif platform.getName() == 'CPU' and prec != 'single':
        warnings.warn('The CPU platform only uses single precision', SimulationWarning)

    # Now get the energy
    has_pbc = parm.ptr('ifbox') > 0
    nrg = u.kilocalories/u.mole
    if decomp:
        # Now we have to loop through every force group and compute them individually

        # Bonds first
        state = context.getState(getEnergy=True, enforcePeriodicBox=has_pbc,
                                 groups=2**parm.BOND_FORCE_GROUP)
        bond = state.getPotentialEnergy().value_in_unit(nrg)
        # Now angles
        state = context.getState(getEnergy=True, enforcePeriodicBox=has_pbc,
                                 groups=2**parm.ANGLE_FORCE_GROUP)
        angle = state.getPotentialEnergy().value_in_unit(nrg)
        # Now dihedrals
        state = context.getState(getEnergy=True, enforcePeriodicBox=has_pbc,
                                 groups=2**parm.DIHEDRAL_FORCE_GROUP)
        dihedral = state.getPotentialEnergy().value_in_unit(nrg)
        # Now the CHARMM-specific terms
        if parm.chamber:
            # Now Urey-Bradley energy
            state = context.getState(getEnergy=True, enforcePeriodicBox=has_pbc,
                                     groups=2**parm.UREY_BRADLEY_FORCE_GROUP)
            ub = state.getPotentialEnergy().value_in_unit(nrg)
            # Now improper
            state = context.getState(getEnergy=True, enforcePeriodicBox=has_pbc,
                                     groups=2**parm.IMPROPER_FORCE_GROUP)
            imp = state.getPotentialEnergy().value_in_unit(nrg)
            # Now cmap
            if parm.has_cmap:
                state = context.getState(getEnergy=True, enforcePeriodicBox=has_pbc,
                                         groups=2**parm.CMAP_FORCE_GROUP)
                cmap = state.getPotentialEnergy().value_in_unit(nrg)
            else:
                cmap = 0

        # Now non-bonded. No real way to decompose this.
        state = context.getState(getEnergy=True, enforcePeriodicBox=has_pbc,
                                 groups=2**parm.NONBONDED_FORCE_GROUP)
        nonbond = state.getPotentialEnergy().value_in_unit(nrg)

        if parm.chamber:
            output.write('Bond         = %20.7f     Angle        = %20.7f\n'
                         'Dihedral     = %20.7f     Urey-Bradley = %20.7f\n'
                         'Improper     = %20.7f     ' % (bond, angle, dihedral, ub, imp))
            if parm.has_cmap:
                output.write('CMAP         = %20.7f\n' % cmap)
            output.write('Nonbond      = %20.7f\nTOTAL        = %20.7f\n' % (nonbond,
                         bond+angle+dihedral+ub+imp+cmap+nonbond))
        else:
            output.write('Bond     = %20.7f     Angle    = %20.7f\n'
                         'Dihedral = %20.7f     Nonbond  = %20.7f\n'
                         'TOTAL    = %20.7f\n' % (bond, angle, dihedral, nonbond,
                         (bond+angle+dihedral+nonbond)))
    else:
        state = context.getState(getEnergy=True, enforcePeriodicBox=parm.ptr('ifbox') > 0)

        output.write('Potential Energy = %.7f kcal/mol\n' %
                     state.getPotentialEnergy().value_in_unit(nrg))

def minimize(parm, igb, saltcon, cutoff, restraintmask, weight,
             script, platform, precision, norun, tol, maxcyc):
    """ Minimizes a snapshot. Use the existing System if it exists """
    global HAS_OPENMM
    if not HAS_OPENMM:
        raise SimulationError('Could not import OpenMM!')

    # Open the script file
    if script is not None:
        scriptfile = open(script, 'w')
        scriptfile.write(_SCRIPT_HEADER % (parm, parm.crdname))
    else:
        scriptfile = None

    gbmeth, kappa = None, 0.0
    if parm.ptr('ifbox') == 0:
        if cutoff is None or cutoff >= 500:
            nbmeth = ff.NoCutoff
            cutoff = 1000.0
        else:
            nbmeth = ff.CutoffNonPeriodic
        if not igb in (0, 1, 2, 5, 6, 7, 8):
            raise SimulationError('Bad igb value. Must be 0, 1, 2, 5, 6, 7, or 8')
        if igb == 1:
            gbmeth = HCT
        elif igb == 2:
            gbmeth = OBC1
        elif igb == 5:
            gbmeth = OBC2
        elif igb == 7:
            gbmeth = GBn
        elif igb == 8:
            gbmeth = GBn2
        # Other choices are vacuum
        kappa = 0.73 * sqrt(saltcon * 0.10806)
    else:
        if cutoff is None: cutoff = 8.0
        nbmeth = ff.PME

    # Time to create the OpenMM system
    system = parm.createSystem(nonbondedMethod=nbmeth, nonbondedCutoff=cutoff*u.angstrom,
                               constraints=None, rigidWater=True, removeCMMotion=False,
                               implicitSolvent=gbmeth, implicitSolventKappa=kappa*(1.0/u.angstrom),
                               soluteDielectric=1.0, solventDielectric=78.5,
                               ewaldErrorTolerance=5e-5)

    if script is not None:
        scriptfile.write('# Create the system\n'
                         'system = parm.createSystem(nonbondedMethod=%s,\n'
                         '                   nonbondedCutoff=%s*u.angstrom,\n'
                         '                   constraints=None, rigidWater=True,\n'
                         '                   removeCMMotion=False, implicitSolvent=%s,\n'
                         '                   implicitSolventKappa=%s*(1.0/u.angstrom),\n'
                         '                   soluteDielectric=1.0,\n'
                         '                   solventDielectric=78.5,\n'
                         '                   ewaldErrorTolerance=5e-5\n'
                         ')\n\n'
                         '# Create a dummy integrator\n'
                         'integrator = VerletIntegrator(2.0*u.femtoseconds)\n\n' %
                         (nbmeth, cutoff, gbmeth, kappa))

    # See if we need to add restraints
    if restraintmask is not None:
        system.addForce(
            positional_restraints(restraintmask, u.kilocalorie_per_mole/u.angstrom/u.angstrom,
                                  parm.coordinates, scriptfile=scriptfile,)
        )

    if script is not None:
        scriptfile.write('system.addForce(frc)\n\n')

    # Create a dummy integrator and the simulation
    integrator = mm.VerletIntegrator(2.0*u.femtoseconds)
    if platform is not None:
        plat = mm.Platform.getPlatformByName(platform)
        simulation = Simulation(parm.topology, system, integrator, plat)
        if script is not None:
            scriptfile.write('# Create the platform\n'
                             'plat = Platform.getPlatformByName(%s)\n\n'
                             '# Create the simulation\n'
                             'simulation = Simulation(parm.topology, system, '
                             'integrator, plat)\n\n' % (platform))
    else:
        simulation = Simulation(parm.topology, system, integrator)
        if script is not None:
            scriptfile.write('# Create the simulation\n'
                             'simulation = Simulation(parm.topology, system, integrator)\n\n')

    # Now assign the coordinates
    simulation.context.setPositions(parm.positions)
    if script is not None:
        scriptfile.write('# Setting the positions\n'
                         'simulation.context.setPositions(parm.positions)\n\n'
                         '# Minimize the energy\n'
                         'simulation.minimizeEnergy(tolerance=%s, maxIterations=%s)\n'
                         '# Store the positions back in the parmtop\n'
                         'state = simulation.context.getState(getPositions=True\n'
                         '                                    enforcePeriodicBox=%s)\n'
                         'parm.positions = state.getPositions()\n' %
                         (bool(parm.ptr('ifbox') > 0), tol, maxcyc or 0))
        scriptfile.close()

    # Go ahead and minimize now and set the coordinates from the results of this
    # minimization if we wanted to run the calculation
    if not norun:
        simulation.minimizeEnergy(tolerance=tol, maxIterations=maxcyc or 0)
        # Now get the coordinates
        state = simulation.context.getState(getPositions=True,
                                            enforcePeriodicBox=parm.ptr('ifbox') > 0)
        parm.positions = state.getPositions()
